"""Container to generate origin showers and also copy relevant files."""

import os
import shutil
import numpy as np
from typing_extensions import Self

from . import write_simulation_to_file, generate_simulation
from smiet import units

import logging

import logging

logger = logging.getLogger("smiet.corsika.generate_origin_library")


class OriginLibraryGenerator:
    corsika_exec_hadr_label : dict = {
        "sibyll": "SIBYLL",
        "qgsjetII": "QGSII",
    }
    """Class to generate the origin shower library using CORSIKA simulations."""

    def __init__(
        self: Self,
        sim_primary: str = "proton",
        sim_energy: float = 10**8 * units.GeV,
        atm_model: int = 17,
        site: str = "lofar",
    ) -> None:
        """
        Initialize the generator.

        Parameters
        ----------
        sim_primary : str, default="proton"
            Primary particle type (e.g., "proton", "iron").

            Currently only "proton" and "iron" are supported.
        sim_energy : float, default=10**8 * units.GeV
            Energy of the primary particle in GeV.
        atm_model : int
            Atmospheric model to use (default is 17).
        site : str
            Site location (default is "lofar") for the magnetic field location
            and the observation level.
        """
        self.sim_primary = sim_primary
        self.sim_energy = sim_energy
        self.atm_model = atm_model
        self.site = site
        self.obs_lev = self.__get_obs_lvl(site)

        self.corsika_exec_name = None
        self.corsika_runner_path = None
        self.run_dir = None

    def __get_obs_lvl(self, site: str) -> float:
        """Set the observation level based on the site."""
        site_obs_levels = {
            "lofar": 0.0 * units.m,
            "auger": 1400.0 * units.m,
        }
        return site_obs_levels[site]

    def setup_paths(
        self: Self,
        corsika_runner_path: str,
        run_dir: str,
        cluster_type: str = "horeka",
        hadr_model: str = "sibyll",
    ) -> None:
        """
        Configure the paths for CORSIKA runner and run directory.

        Parameters
        ----------
        corsika_runner_path : str
            Path to the CORSIKA runner executable.
        run_dir : str
            Directory where the CORSIKA simulations will be run and saved.
        cluster_type : str, default="horeka"
            Type of cluster that the jobs are submitted into.

            The submit scripts should be placed in the submit_templates directory,
            with the naming scheme submit_{cluster_type}.sh.

            Currently only "horeka" is supported.
        hadr_model : str, default="sibyll"
            Hadronic interaction model used for the fitting functions.
            Currently only "sibyll" and "qgsjetII" are supported.
        """
        self.corsika_runner_path = corsika_runner_path
        self.run_dir = run_dir

        # ensure that the directories exist
        if not os.path.exists(self.corsika_runner_path):
            raise FileNotFoundError(
                f"CORSIKA runner path {self.corsika_runner_path} not found."
            )
        if not os.path.exists(self.run_dir):
            os.makedirs(self.run_dir, exist_ok=True)

        self.hadr_model = hadr_model
        # name of coreas runner executable
        self.corsika_exec_name = f"mpi_corsika77550Linux_{self.corsika_exec_hadr_label[hadr_model]}_urqmd_thin_coreas_parallel_runner"

        # ensure that the corsika executable exists
        if not os.path.exists(
            os.path.join(corsika_runner_path, self.corsika_exec_name)
        ):
            print(
                f"CORSIKA executable {self.corsika_exec_name} not found in {corsika_runner_path}. Exiting."
            )
            raise FileNotFoundError()

        path_to_submit_tmpl = os.path.join(
            os.path.abspath(os.path.dirname(__file__)),
            "submit_templates",
            f"submit_{cluster_type}.sh",
        )

        # make sure it exists
        if not os.path.exists(path_to_submit_tmpl):
            raise FileNotFoundError(
                f"Submit template for cluster type {cluster_type} not found at {path_to_submit_tmpl}."
            )

        with open(path_to_submit_tmpl, "r") as f:
            self.submit_file_templ = f.read()

    def generate_submit_files(
        self: Self,
        zenith_angles: tuple = (0, 63, 3),
        xmax_values: tuple = (600, 1300, 100),
        n_sims: int = 3,
        n_nodes: int = 2,
        n_cores_per_node: int = 76,
        vxvxB: bool = True,
    ) -> list:
        """
        Generate steering files for CORSIKA simulations.

        Parameters
        ----------
        zenith_angles : tuple
            Tuple specifying (min, max, step) for zenith angles in degrees.
            NOte that max should be less than 60 degrees, and include the step.
        xmax_values : tuple
            Tuple specifying (min, max, step) for Xmax values in g/cm^2.
            NOte that max should include the step.
        n_sims : int, default=3
            The number of times to repeat each simulation (default is 3).
            This is to ensure that we have enough statistics.
        n_nodes : int
            Number of nodes to use for each simulation (default is 2).
        n_cores_per_node : int
            Number of cores per node (default is 76).
        vxvxB : bool
            Whether to only generate showers on the vxvxB axis or not.
            Default is True. Otherwise, a rotated starshape is used.

        Returns
        -------
        list
            List of directories where simulations are set up.
        """
        if vxvxB:
            nr_of_arms = 1
            rotate = False
        else:
            nr_of_arms = 8
            rotate = True
        sim_cores = int(n_nodes * n_cores_per_node)

        zeniths = np.arange(*zenith_angles)
        xmaxs = np.arange(*xmax_values)

        sim_rundirs = []

        # check if the run directory has been set up
        if self.run_dir is None:
            raise ValueError(
                "Run directory not set up. Please run setup_paths() first."
            )

        # for bookkeeping, generate a file that takes in the simulation number, zenith, and xmax
        sim_labels = ["SIM\tZENITH(deg)\tXMAX(g/cm2)\tITER"]

        for izen, ixmax, isim in np.ndindex(len(zeniths), len(xmaxs), n_sims):

            sim_zenith = zeniths[izen] * units.deg
            sim_xmax = xmaxs[ixmax]
            # compute the simulation number based on a basic criteria
            sim_nr = int(f"00{izen:02d}{isim:01d}{ixmax:01d}")

            inp_file, lst_file, reas_file = generate_simulation(
                sim_nr=sim_nr,
                sim_primary=self.sim_primary,
                sim_cores=sim_cores,
                sim_zenith=sim_zenith,
                sim_energy=self.sim_energy,
                sim_xmax=sim_xmax,
                sim_azimuth=0.0 * units.deg,  # fixed here
                slice_gram=5.0,  # fixed here
                magnetic_field=self.site,
                atmosphere=self.atm_model,
                nr_of_arms=nr_of_arms,
                rotate=rotate,
                core=np.array([0, 0, self.obs_lev]),
            )

            # create the necessary directories
            sim_dir = os.path.join(
                self.run_dir, f"zen{sim_zenith / units.deg:.0f}", f"xmax{sim_xmax:.0f}"
            )
            sim_rundir = os.path.join(sim_dir, f"SIM{sim_nr:06d}")
            os.makedirs(sim_dir, exist_ok=True)

            # check if a simulation has already been ran, if so then skip
            if os.path.exists(os.path.join(sim_rundir, f"DAT{sim_nr:06d}.long")):
                logger.warning(f"SIM{sim_nr:06d} already ran. Skipping.")
                continue
            # manually modify the input file to add the direct and datdir inputs
            inp_file = self.modify_input_file(inp_file, sim_rundir)

            # write to directory
            write_simulation_to_file(
                inp_file, lst_file, reas_file, sim_directory=sim_dir, reset=True
            )

            # create a logs directory
            os.makedirs(os.path.join(sim_dir, "logs"), exist_ok=True)

            # copy the coreas_to_hdf5_origin.py script to the directory
            shutil.copy(
                os.path.join(os.path.dirname(__file__), "coreas_to_hdf5_origin.py"),
                os.path.join(self.run_dir, "coreas_to_hdf5_origin.py"),
            )

            # also create a hdf5 directory to store the hdf5 files
            os.makedirs(os.path.join(self.run_dir, "hdf5_sims"), exist_ok=True)

            # also create the submit file
            submit_file = self.submit_file_templ.format(
                sim_xmax,
                sim_zenith / units.deg,
                sim_dir,
                n_nodes,
                n_cores_per_node,
                self.corsika_runner_path,
                self.corsika_exec_name,
                sim_nr,
                self.run_dir,
            )
            with open(os.path.join(sim_rundir, "submit.sh"), "w") as f:
                f.write(submit_file)

            sim_labels.append(
                f"{sim_nr:06d}\t{sim_zenith / units.deg:.1f}\t{sim_xmax:.1f}\t{isim+1}"
            )
            sim_rundirs.append(sim_rundir)

        logger.info(
            f"Generated {len(sim_rundirs)} simulation directories in {self.run_dir}"
        )

        # write the sim_labels to a file
        with open(os.path.join(self.run_dir, "sim_labels.txt"), "w") as f:
            f.write("\n".join(sim_labels))

        return sim_rundirs

    def submit_jobs(self: Self, sim_rundirs: list) -> None:
        """
        Submit the generated jobs to the SLURM scheduler.

        Parameters
        ----------
        sim_rundirs : list
            List of directories where simulations are set up.
        """
        for sim_rundir in sim_rundirs:
            submit_cmd = f"sbatch {os.path.join(sim_rundir, 'submit.sh')}"
            logger.info(f"Submitting job with command: {submit_cmd}")
            os.system(submit_cmd)

    def modify_input_file(self: Self, inp_file: list, direct: str) -> list:
        """
        Modify the input file to include direct and datdir commands.

        Parameters
        ----------
        inp_file : list
            List of strings representing the lines of the input file.

        Returns
        -------
        list
            Modified list of strings for the input file.
        """
        modified_inp = []
        for line in inp_file:
            if line.startswith("DIRECT"):
                modified_inp.append(f"DATDIR {self.corsika_runner_path}\n")
                modified_inp.append(f"DIRECT {direct}/\n")
            else:
                modified_inp.append(line)
        return modified_inp

    def generate_single_shower(
        self: Self,
        sim_zenith: float,
        sim_xmax: float,
        n_nodes: int = 2,
        n_cores_per_node: int = 76,
        vxvxB: bool = True,
        sim_dir="./",
        sim_nr: int = 999999,
    ) -> str:
        """
        Generate a single origin shower simulation. This is useful for testing and debugging purposes.

        Parameters
        ----------
        sim_zenith : float
            Zenith angle in degrees.
        sim_xmax : float
            Xmax value in g/cm^2.
         n_nodes : int
            Number of nodes to use for each simulation (default is 2).
        n_cores_per_node : int
            Number of cores per node (default is 76).
        vxvxB : bool
            Whether to only generate showers on the vxvxB axis or not.
            Default is True. Otherwise, a rotate starshape is used.
        sim_dir : str
            Directory where the simulation will be set up.
        sim_nr : int, default=999999
            Simulation number to use for the shower.
        """
        if vxvxB:
            nr_of_arms = 1
            rotate = False
        else:
            nr_of_arms = 8
            rotate = True
        inp_file, lst_file, reas_file = generate_simulation(
            sim_nr=sim_nr,
            sim_primary=self.sim_primary,
            sim_cores=int(n_nodes * n_cores_per_node),
            sim_zenith=sim_zenith * units.deg,
            sim_energy=self.sim_energy,
            sim_xmax=sim_xmax,
            slice_gram=5.0,
            magnetic_field=self.site,
            atmosphere=self.atm_model,
            nr_of_arms=nr_of_arms,
            rotate=rotate,
            core=np.array([0, 0, self.obs_lev]),
        )

        # manually modify the input file to add the direct and datdir inputs
        inp_file = self.modify_input_file(inp_file, sim_dir)

        write_simulation_to_file(
            inp_file, lst_file, reas_file, sim_directory=sim_dir, reset=True
        )

        return inp_file, lst_file, reas_file



import os

import numpy as np
import radiotools.helper as hp

from radiotools.coordinatesystems import cstrafo

from smiet import units

import logging

logger = logging.getLogger("smiet.corsika.generate_file_contents")


# Functions to generate the individual files


def generate_inp_file(
    sim_nr,
    sim_zenith,
    sim_azimuth,
    sim_energy,
    primary=14,
    direct="./",
    magnet=(0.0, 0.186 * units.gauss, -0.456 * units.gauss),
    obs_lev=0.0 * units.m,
    slicing=5,
    first_height = None,
    thin=1e-7,
    n_cores=10,
    atm=17,
):
    """
    Make an INP file for a CORSIKA run with CoREAS, using MPI.

    As CoREAS only supports a single shower per run, the `EVTNR` and `NSHOW` are set to 1.
    The shower zenith, azimuth and energy are fixed to the values provided.
    For the geometry, the NRR coordinate system is assumed (ie x-axis points east), which means
    that the provided azimuth is shifted by 90 degrees to convert to the CORSIKA system.
    The SEED values are randomly generated. For the USER variable, the login name is used (by
    evaluating `os.getlogin()`).

    Parameters
    ----------
    sim_nr : int
        The value for the RUNNR parameter
    sim_zenith : float
        The zenith of the shower (THETAP). Currently only using a fixed zenith is supported
    sim_azimuth : float
        The azimuth of the shower (PHIP + 270 deg). Currently only using a fixed azimuth is supported
    sim_energy : float
        The fixed energy to use for the simulation
    primary : int, default=14
        The CORSIKA identifier for the primary particle type. Default is 14 (proton).
    obs_lev : float, default=0.0 * units.m
        The observation level to use for the simulation (OBSLEV)
    magnet : tuple of float, default=(0.0, 18.6, -45.6) * units.gauss
        The components of the 3D magnetic field, provided in the NRR CS and internal units (MAGNET)
    first_height : float or None, default=None
        The height of the first interaction, passed into the FIXHEI parameter.
        If not None, then the units must be in internal units. 
        If None, then this parameter is not set.
    direct : str, default='./'
        The directory for the particle output (DIRECT)
    atm : int or str, default=17
        If an integer, the CORSIKA identifier for the atmosphere (used for ATMOD).
        If a string, the name of the GDAS atmosphere file (used for ATMFILE)
    n_cores : int, default=10
        The number of cores over which the MPI simulation will run.
        If n_cores = 1, the PARALLEL option will be disabled and a simulation without MPI will run.
    thin : float, default=1e-7
        The thinning level to use (THIN)
    slicing : float, default=5
        The thickness of the atmospheric slices, to be used for the longitudinal profile (LONGI)
        Defaults to units of g/cm^2.

    Returns
    -------
    contents : list of str
        The contents to be written to the INP file
    """
    atm_line = f"ATMFILE {atm}\n" if type(atm) is str else f"ATMOD   {atm}\n"
    contents = [
        # Add run information
        f"RUNNR   {sim_nr}\n",
        f"EVTNR   1\n",
        f"NSHOW   1\n",
        f"{atm_line}",
        # Set SEED with random numbers (MPI option uses 6 seeds)
        f"SEED    {np.random.randint(1, 900000001)} 0 0\n",
        f"SEED    {np.random.randint(1, 900000001)} 0 0\n",
        f"SEED    {np.random.randint(1, 900000001)} 0 0\n",
        f"SEED    {np.random.randint(1, 900000001)} 0 0\n",
        f"SEED    {np.random.randint(1, 900000001)} 0 0\n",
        f"SEED    {np.random.randint(1, 900000001)} 0 0\n",
        # Set primary particle
        f"PRMPAR  {primary}\n",
        # Set fixed energy
        f"ERANGE  {sim_energy / units.GeV:.2f} {sim_energy / units.GeV:.2f}\n",
        f"ESLOPE  0.000E+00\n",
        f"ECUTS   3.000E-01 3.000E-01 4.010E-04 4.010E-04\n",
        # Add thinning, coupled to ERANGE
        f"THIN    {thin:.2e} {sim_energy / units.GeV * thin:.2e} 0.000E+00\n",
        f"THINH   1.000E+00 1.000E+02\n",
        # Control geometry
        f"THETAP  {sim_zenith / units.deg:.1f} {sim_zenith / units.deg:.1f}\n",
        f"PHIP    {-270.0 + sim_azimuth / units.deg:.1f} {-270.0 + sim_azimuth / units.deg:.1f}\n",
        f"MAGNET  {magnet[1] / units.microtesla:.2f} {-magnet[2] / units.microtesla:.2f}\n",
        f"OBSLEV  {obs_lev / units.cm:.2f}\n",
        # Electromagnetic interactions
        f"ELMFLG  T T\n",
        f"STEPFC  1.000E+00\n",
        f"RADNKG  5.000E+05\n",
        f"MUMULT  T\n",
        f"MUADDI  T\n",
        f"MAXPRT  1\n",
        f"ECTMAP  1.000E+5\n",
        # Particle output
        f"PAROUT  F F\n",
        f"DATBAS  F\n",
        f"LONGI   T {slicing:.1f} T T\n",
    ]
    # if first_height is not None, then set FIXHEI parameter
    if first_height is not None:
        contents.append(f"FIXHEI  {first_height / units.cm:.2f} 0\n")
    # only add PARALLEL option if n_cores > 1
    # in this way, it can also be used to generate normal CORSIKA simulations (maybe)
    if n_cores > 1:
        # Coupling ERANGE with PARALLEL
        contents.append(
            f"PARALLEL 1000 {sim_energy / units.GeV / n_cores / 100:.1e} 1 T\n"
        )
    else:
        logger.warning(
            "Number of cores is set to 1. Disabling PARALLEL option, i.e. running CORSIKA without MPI. This is not recommended for generating origin showers."
        )

    contents += [
        # User information
        f"USER    {os.getlogin()}\n",
        f"DIRECT  {direct}\n",
        # End the steering file
        f"EXIT\n",
    ]

    return contents


def generate_list_file(
    sim_zenith,
    sim_azimuth,
    slices_gram,
    magnet_vector=None,
    number_of_arms=8,
    core=None,
    radii=None,
    rotate : bool = False
):
    """
    Create a starshape pattern with 8 arms in the shower plane and project it to the ground along the shower axis.

    This result in the antennas from the same "circle" to have different
    viewing angles after being projected to the ground. The antennas are sliced according to
    the grammages specified, and in the LIST file the receivers are ordered by slice (which
    proves to be a couple of percent faster than ordering them by position).

    Parameters
    ----------
    sim_zenith : float
        The zenith angle of the shower, in the NRR coordinate system
    sim_azimuth : float
        The azimuth angle of the shower, in the NRR coordinate system
    slices_gram : list of float
        The grammages of the planes defining the slices, including both the first and last one.
    core : list of float, default=[0,0,0] * units.m
        The core of the simulated air shower
    magnet_vector : array of float, default=[0, 0.186, -0.456] * units.gauss
        The magnetic field vector used in the simulation, used to project the antennas from
        the shower plane to the ground
    radii : list of float, default=None
        The distances (in the showerplane) from the showeraxis on which to put the antennas
        If None, a default starshape is used.
    number_of_arms: int, default=8
        The number of arms to use for the star shape pattern
    rotate : bool, default=False
        Whether to rotate the starshape pattern by 22.5 degrees, to avoid alignment of antennas
        along the vxB axis, which shows to fail with template synthesis.

    Returns
    -------
    contents : list of str
        The lines to be written to the LIST file
    """
    if magnet_vector is None:
        magnet_vector = np.array([0.0, 0.186, -0.456]) * units.gauss
    if core is None:
        core = np.array([0.0, 0.0, 0.0])
    if radii is None:
        radii = (
            np.concatenate(
                (
                    [12.5 / 8, 12.5 / 4, 12.5 / 2],
                    np.arange(12.5, 200, 12.5),
                    np.arange(200, 300, 25),
                    np.arange(300, 400, 50),
                    np.arange(400, 501, 100),
                )
            )
            * units.m
        )

    ant_positions = _calculate_antenna_positions(
        sim_zenith, sim_azimuth, radii, number_of_arms, magnet_vector, core, rotate
    )

    return generate_list_file_from_positions(ant_positions, slices_gram)


def _calculate_antenna_positions(
    sim_zenith, sim_azimuth, radii, number_of_arms, magnet_vector, core, rotate=False
):
    if len(radii.shape) != 1:
        raise ValueError("Radii should be a 1D array")

    # Make coordinate transformer, taking into account transformation to radiotools CS
    transformer = cstrafo(
        sim_zenith / units.rad,
        sim_azimuth / units.rad,
        magnetic_field_vector=magnet_vector / units.gauss,
    )

    logger.info(f"Number of antennas to simulate: {len(radii) * number_of_arms:d}")

    if rotate:
        logger.info("Rotating starshape pattern by 22.5 degrees")
        ang_shift = 22.5 * units.deg
    else:
        ang_shift = 0.0 * units.deg

    ant_position_vvB = np.zeros((len(radii) * number_of_arms, 3), dtype=float)
    radians_step = 2 * np.pi / number_of_arms
    for ind, r in enumerate(radii):
        for j in np.arange(number_of_arms):
            # v = 0
            vB = r * np.sin(j * radians_step + ang_shift / units.rad)
            vvB = r * np.cos(j * radians_step + ang_shift / units.rad)

            ant_position_vvB[int(ind * number_of_arms + j), :2] = [vB, vvB]

    ant_position = transformer.transform_from_vxB_vxvxB_2D(
        ant_position_vvB, core=core / units.m
    )

    ant_position /= units.cm  # put the positions in cm to write them to LIST file

    return ant_position


def generate_list_file_from_positions(ant_positions, slices_gram):
    """
    Generate a list file from a known antenna layout from existing simulations.

    The function simply adds one observer for every antenna position in `ant_positions`, for each
    slice in `slices_gram`. This can be used to generate list files from simulations that already
    exist for the sake of generating an origin shower with the same layout.

    Parameters
    ----------
    ant_positions : list of [list of float]
        list of antenna positions for each CoREAS observer.
    slices_gram : list of float
        The grammages of the planes defining the slices, including both the first and last one.

    Returns
    -------
    contents : list of str
        The lines to be written to the LIST file
    """
    contents = []

    logger.info(f"Number of slices to simulate: {len(slices_gram)}")

    for slice_ind in range(len(slices_gram) - 1):
        for idx, ant in enumerate(ant_positions):
            ant_name = (
                f"{int(ant[1] >= 0)}{int(abs(ant[1]) / 100)}_"
                f"{int(ant[0] <= 0)}{int(abs(ant[0]) / 100)}_{idx}"
            )

            ant_line = "AntennaPosition = "
            ant_line += f"{ant[1]:.1f} {-1 * ant[0]:.1f} {ant[2]:.1f} "  # here taking the values DIRECTLY, so NO TRANSFORM
            ant_line += ant_name + f"x{round(slices_gram[slice_ind + 1])} "
            ant_line += f"slantdepth {slices_gram[slice_ind]:.1f} {slices_gram[slice_ind + 1]:.1f} "
            ant_line += "\n"

            contents.append(ant_line)
        contents.append("\n")

    return contents


def generate_list_file_cherenkov(
    sim_zenith,
    sim_azimuth,
    slices_gram,
    atm,
    viewing_angle_fractions,
    number_of_arms=8,
    magnet_vector=None,
    core=None,
    rotate : bool = False
):
    """
    Create a starshape pattern with 8 arms in the shower plane and project it to the ground,
    along the **viewing angle** under which antenna observes the slice. Note that this results

    Note that this results
    in different physical antennas on the ground for every slice! Therefore, this function
    should not be used to generate list files for origin showers. The antennas are sliced
    according to the grammages specified, and in the LIST file the receivers are ordered
    by slice (which proves to be a couple of percent faster than ordering them by position).

    Parameters
    ----------
    sim_zenith : float
        The zenith angle of the shower, in the NRR coordinate system
    sim_azimuth : float
        The azimuth angle of the shower, in the NRR coordinate system
    slices_gram : list of float
        The grammages of the planes defining the slices, including both the first and last one.
    atm : radiotools.atmosphere.models.Atmosphere
        The atmosphere model to use to map the antennas to the ground plane
    core : list of float, default=[0,0,0] * units.m
        The core of the simulated air shower
    magnet_vector : array of float, default=[0, 0.186, -0.456] * units.gauss
        The magnetic field vector used in the simulation, used to project the antennas from
        the shower plane to the ground
    viewing_angle_fractions : list of float
        The viewing angles (expressed as fractions of the local Cherenkov angle) under which to
        place the antennas.
    number_of_arms: int, default=8
        The number of arms to use for the star shape pattern
    rotate : bool, default=False
        Whether to rotate the starshape pattern by 22.5 degrees, to avoid alignment of antennas
        on the vxB axis.

    Returns
    -------
    contents : list of str
        The lines to be written to the LIST file
    """
    if magnet_vector is None:
        magnet_vector = np.array([0.0, 0.186, -0.456]) * units.gauss
    if core is None:
        core = np.array([0.0, 0.0, 0.0])
    if viewing_angle_fractions is None:
        raise ValueError("Viewing angle fractions should be provided")
    
    if rotate:
        logger.info("Rotating starshape pattern by 22.5 degrees")
        ang_shift = 22.5 * units.deg
    else:
        ang_shift = 0.0 * units.deg

    # Make radii per slice
    nr_of_slices = len(slices_gram) - 1

    # Create the contents
    contents = []

    # Make coordinate transformer, taking into account transformation to radiotools CS
    transformer = cstrafo(
        sim_zenith / units.rad,
        sim_azimuth / units.rad,
        magnetic_field_vector=magnet_vector / units.gauss,
    )

    for slice_ind in range(nr_of_slices):
        xslice = slices_gram[slice_ind + 1]
        # NB: in radiotools, g/cm^2 is set to 1 and units are converted internally.
        # Therefore, no need to convert xslice here.
        height_slice = max(atm.get_vertical_height(sim_zenith / units.rad, xslice), 0.0)
        cherenkov_slice = np.arccos(1 / atm.get_n(height_slice))
        viewing_angles = cherenkov_slice * viewing_angle_fractions

        d_xmax = (
            atm.get_distance_xmax_geometric(
                sim_zenith / units.rad, xslice, observation_level=0.0
            )
            * units.m
        )

        radii = d_xmax * np.tan(viewing_angles)

        ant_position_vvB = np.zeros(
            (len(viewing_angle_fractions) * number_of_arms, 3), dtype=float
        )
        radians_step = 2 * np.pi / number_of_arms
        for ind, r in enumerate(radii):
            for j in np.arange(number_of_arms):
                # v = 0
                vB = r * np.cos(j * radians_step + ang_shift/units.rad)
                vvB = r * np.sin(j * radians_step + ang_shift/units.rad)

                ant_position_vvB[int(ind * number_of_arms + j), :2] = [vB, vvB]

        ant_position_ground = transformer.transform_from_vxB_vxvxB(
            ant_position_vvB, core=core / units.m
        )

        unit_shower_axis = hp.spherical_to_cartesian(
            sim_zenith / units.rad, sim_azimuth / units.rad
        )
        slice_vector = unit_shower_axis * d_xmax

        delta = ant_position_ground - slice_vector
        alpha = -ant_position_ground[:, 2] / delta[:, 2]

        ant_projected_ground = delta.T * alpha + ant_position_ground.T
        ant_projected_ground /= (
            units.cm
        )  # put the positions in cm to write them to LIST file

        for idx, ant in enumerate(ant_projected_ground.T):
            ant_name = (
                f"{int(ant[1] >= 0)}{int(abs(ant[1]) / 100)}_"
                f"{int(ant[0] <= 0)}{int(abs(ant[0]) / 100)}_{idx}"
            )

            ant_line = "AntennaPosition = "
            ant_line += f"{ant[1]:.1f} {-1 * ant[0]:.1f} {ant[2]:.1f} "
            ant_line += ant_name + f"x{round(slices_gram[slice_ind + 1])} "
            ant_line += f"slantdepth {slices_gram[slice_ind]:.1f} {slices_gram[slice_ind + 1]:.1f} "
            ant_line += "\n"

            contents.append(ant_line)
        contents.append("\n")

    return contents


def generate_reas_file(
    event_number,
    n=1.000292,
    time_res=2e-10 * units.s,
    time_bound=4e-7 * units.s,
    core=None,
):
    """
    Generate the contents of the REAS file using the set of given input parameters.

    The contents are based on the CoREAS V1.4 software.

    Parameters
    ----------
    event_number : int
        The run number of the simulation
    n : float, default=1.000292
        The refractive index at sea level (GroundLevelRefractiveIndex)
    time_res : float, default=2e-10 * units.s
        The time resolution to use for the CoREAS simulation (TimeResolution)
    time_bound : float, default=4e-7 * units.s
        The time window to use for the CoREAS simulation (used with AutomaticTimeBoundaries)
    core : list of float, default=[0,0,0] * units.m
        The core of the simulated air shower, in the NRR coordinate system

    Returns
    -------
    contents : list of str
        The lines to be written to the REAS file
    """
    if core is None:
        core = [0.0, 0.0, 0.0]

    corsika_core = [core[1], -1 * core[0], core[2]]

    contents = [
        "# CoREAS V1.4 by Tim Huege <tim.huege@kit.edu> with contributions by Marianne Ludwig and Clancy James - parameter file\n"
        "\n",
        "# parameters setting up the spatial observer configuration:",
        "\n",
        f"CoreCoordinateNorth = {corsika_core[0] / units.cm:.1f}     ; in cm\n",
        f"CoreCoordinateWest = {corsika_core[1] / units.cm:.1f}     ; in cm\n",
        f"CoreCoordinateVertical = {corsika_core[2] / units.cm:.1f}     ; in cm\n",
        "\n",
        "# parameters setting up the temporal observer configuration:\n",
        "\n",
        f"TimeResolution = {time_res / units.s:.2e}  				; in s\n",
        f"AutomaticTimeBoundaries = {time_bound / units.s:.2e}			; 0: off, x: automatic boundaries with width x in s\n",
        "TimeLowerBoundary = -1				; in s, only if AutomaticTimeBoundaries set to 0\n",
        "TimeUpperBoundary = 1				; in s, only if AutomaticTimeBoundaries set to 0\n",
        "ResolutionReductionScale = 0			; 0: off, x: decrease time resolution linearly every x cm in radius\n",
        "\n",
        "# parameters setting up the simulation functionality:\n",
        "\n",
        "GroundLevelRefractiveIndex = %.8f        ; specify refractive index at 0 m asl\n"
        % n,
        "\n",
        "# event information for Offline simulations:\n",
        "\n",
        "EventNumber = -1\n",
        "RunNumber = -1\n",
        "GPSSecs = 0\n",
        "GPSNanoSecs = 0\n",
        "CoreEastingOffline = 0.0                ; in meters\n",
        "CoreNorthingOffline = 0.0                ; in meters\n",
        "CoreVerticalOffline = 0.0                ; in meters\n",
        "OfflineCoordinateSystem = Reference\n",
        "RotationAngleForMagfieldDeclination = 0.0        ; in degrees\n",
        "Comment =\n",
        "\n",
        "# event information for your convenience and backwards compatibility with other software, these values are not used as input parameters for the simulation:"
        "\n"
        "CorsikaFilePath = ./\n",
        f"CorsikaParameterFile = SIM{event_number:06d}.inp\n",
    ]

    return contents


"""
Utility modules required for JAX implementation.

# Copyright(C) 2022 Gordian Edenhofer
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause
# Authors: Gordian Edenhofer, Philipp Frank
"""

import operator
from abc import ABCMeta
from dataclasses import dataclass
from functools import partial
from typing_extensions import Union

import jax
from jax.tree_util import register_pytree_node, register_pytree_node_class, tree_map

"""
Below code is used to define the metaclass that is used for all subsequent definitions
of classes used in this module. It properly takes care of extending PyTree definitions
to 

Obtained from ift/NIFTy/src/re/model.py

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause 
"""


class ModelMeta(ABCMeta):
    """Register all derived classes as PyTrees in JAX using metaprogramming.

    For any dataclasses.Field property with a metadata-entry named "static",
    we will either hide or expose the property to JAX depending on the value.
    """

    def __new__(mcs, name, bases, dict_, /, **kwargs):
        cls = super().__new__(mcs, name, bases, dict_, **kwargs)
        cls = dataclass(init=False, repr=False, eq=False)(cls)
        IS_STATIC_DEFAULT = True

        def tree_flatten(self):
            static = []
            dynamic = []
            for k, v in self.__dict__.items():
                # Inspired by how equinox registers properties as static in JAX
                fm = self.__dataclass_fields__.get(k)
                fm = fm.metadata if fm is not None else {}
                if fm.get("static", IS_STATIC_DEFAULT) is False:
                    dynamic.append((PyTreeString(k), v))
                else:
                    static.append((k, v))
            return (tuple(dynamic), tuple(static))

        @partial(partial, cls=cls)
        def tree_unflatten(aux, children, *, cls):
            static, dynamic = aux, children
            obj = object.__new__(cls)
            for nm, m in dynamic + static:
                setattr(obj, str(nm), m)  # unwrap any potential `PyTreeSring`s
            return obj

        # Register class and all classes deriving from it
        register_pytree_node(cls, tree_flatten, tree_unflatten)
        return cls


"""
Below code is used to defined PyTreeString objects, which defines strings
as PyTree objects that can be propagated as a dynamic object in the PyTree
structure.

Obtained from ift/NIFTy/src/re/tree_math/pytree_string.py

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause 
"""


def _unary_op(op, name=None):
    def unary_call(lhs):
        return op(lhs._str)

    name = op.__name__ if name is None else name
    unary_call.__name__ = f"__{name}__"
    return unary_call


def _binary_op(op, name=None):
    def binary_call(lhs, rhs):
        lhs = lhs._str if isinstance(lhs, PyTreeString) else lhs
        rhs = rhs._str if isinstance(rhs, PyTreeString) else rhs
        out = op(lhs, rhs)
        return PyTreeString(out) if isinstance(out, str) else out

    name = op.__name__ if name is None else name
    binary_call.__name__ = f"__{name}__"
    return binary_call


def _rev_binary_op(op, name=None):
    def binary_call(lhs, rhs):
        lhs = lhs._str if isinstance(lhs, PyTreeString) else lhs
        rhs = rhs._str if isinstance(rhs, PyTreeString) else rhs
        out = op(rhs, lhs)
        return PyTreeString(out) if isinstance(out, str) else out

    name = op.__name__ if name is None else name
    binary_call.__name__ = f"__r{name}__"
    return binary_call


def _fwd_rev_binary_op(op, name=None):
    return (_binary_op(op, name=name), _rev_binary_op(op, name=name))


@register_pytree_node_class
class PyTreeString:
    def __init__(self, str):
        self._str = str

    def tree_flatten(self):
        return ((), (self._str,))

    @classmethod
    def tree_unflatten(cls, aux, _):
        return cls(*aux)

    def __str__(self):
        return self._str

    def __repr__(self):
        return f"{self.__class__.__name__}({self._str!r})"

    __lt__ = _binary_op(operator.lt)
    __le__ = _binary_op(operator.le)
    __eq__ = _binary_op(operator.eq)
    __ne__ = _binary_op(operator.ne)
    __ge__ = _binary_op(operator.ge)
    __gt__ = _binary_op(operator.gt)

    __add__, __radd__ = _fwd_rev_binary_op(operator.add)
    __mul__, __rmul__ = _fwd_rev_binary_op(operator.mul)

    lower = _unary_op(str.lower)
    upper = _unary_op(str.upper)

    __hash__ = _unary_op(str.__hash__)

    startswith = _binary_op(str.startswith)


def hide_strings(a):
    return tree_map(lambda x: PyTreeString(x) if isinstance(x, str) else x, a)


def jax_zip(x: jax.typing.ArrayLike, y: jax.typing.ArrayLike) -> list[tuple]:
    """equivalent of zip function in Python"""

    assert x.shape[0] == y.shape[0], "lengths do not match."

    return tree_map(
        lambda xx, yy: jax.numpy.array([[xx[i], yy[i]] for i in range(x.shape[0])]),
        x,
        y,
    )


def jit_repeat(
    a: jax.typing.ArrayLike,
    repeats: jax.typing.ArrayLike,
    axis: Union[int, None],
    total_repeat_length: Union[int, None],
):
    """Jit-compiled version of jax.numpy.repeat."""
    return jax.jit(jax.numpy.repeat, static_argnames=["axis", "total_repeat_length"])(
        a, repeats, axis, total_repeat_length
    )



import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import radiotools.helper as hp

from smiet import units
from smiet.numpy import SlicedShowerCherenkov, SlicedShower, TemplateSynthesis
from smiet.numpy.utilities import angle_between


def calculate_distance_to_ant(the_slice, shower, ant):
    d_xmax = the_slice._atm.get_distance_xmax_geometric(
        shower.zenith / units.rad, the_slice.grammage, observation_level=shower.antenna_dict[ant][2]
    ) * units.m

    unit_shower_axis = hp.spherical_to_cartesian(shower.zenith / units.rad, shower.azimuth / units.rad)
    slice_vector = unit_shower_axis * d_xmax

    ant_pos = np.array([-1 * shower.antenna_dict[ant][1], shower.antenna_dict[ant][0], shower.antenna_dict[ant][2]])
    slice_to_ant_vector = ant_pos - slice_vector

    return np.linalg.norm(slice_to_ant_vector, axis=-1)


def calculate_viewing_angle(the_slice, shower, ant):
    d_xmax = the_slice._atm.get_distance_xmax_geometric(
        shower.zenith / units.rad, the_slice.grammage, observation_level=shower.antenna_dict[ant][2]
    ) * units.m

    unit_shower_axis = hp.spherical_to_cartesian(shower.zenith / units.rad, shower.azimuth / units.rad)
    slice_vector = unit_shower_axis * d_xmax

    ant_pos = np.array([-1 * shower.antenna_dict[ant][1], shower.antenna_dict[ant][0], shower.antenna_dict[ant][2]])
    slice_to_ant_vector = ant_pos - slice_vector

    cherenkov_angle = the_slice._calculate_cherenkov_angle(shower.zenith)

    return angle_between(slice_to_ant_vector, -1 * slice_vector) / cherenkov_angle


def calculate_viewing_angle_ant(the_slice, shower, ant):
    from radiotools.atmosphere.cherenkov_radius import get_cherenkov_angle
    ant_pos = shower.get_ant_position_showerplane(ant)

    viewing_angle = my_synthesis.atm.get_viewing_angle(
        shower.zenith / units.rad, np.linalg.norm(ant_pos[:2]),
        xmax=the_slice.grammage, observation_level=shower.antenna_dict[ant][2]
    )
    cherenkov_angle = get_cherenkov_angle(
        the_slice._atm.get_vertical_height(shower.zenith / units.rad, the_slice.grammage), (1 + 292e-6), 17
    )

    return viewing_angle / cherenkov_angle


GRAMMAGE = 900
# COPIED FROM GEN FILE
VIEWING_ANGLES = np.array([
    0.001, 0.004, 0.007,
    0.01, 0.05, 0.10, 0.50,
    0.97, 0.98, 0.99, 1.00,
    1.01, 1.02, 1.03, 1.50,
    2.00, 2.50, 3.00, 4.00,
    5.00, 6.00, 7.00, 8.00,
    9.00, 10.0,
])


if __name__ == '__main__':
    origin = SlicedShower('/home/mitjadesmet/Data/Showers/SIM130057.hdf5')
    my_shower = SlicedShowerCherenkov('/home/mitjadesmet/Data/Showers/SIM160000.hdf5')
    my_shower_2 = SlicedShowerCherenkov('/home/mitjadesmet/Data/Showers/SIM160001.hdf5')

    my_synthesis = TemplateSynthesis(freq_ar=[30, 500, 100])
    my_synthesis.make_template(origin)

    my_slice = my_synthesis.slices[GRAMMAGE // 5 - 1]

    matching_index_viewing_angles = {
        angle: np.abs(angle - my_synthesis.viewing_angles).argmin()
        for angle in VIEWING_ANGLES
    }

    my_ant = np.array([el for el in my_shower.antenna_names if el.split('x')[1] == f'{GRAMMAGE}'])
    my_pos = np.array([my_shower.get_ant_position_showerplane(ant) for ant in my_ant])
    my_dist = np.sqrt(np.sum(my_pos[:, :2] ** 2, axis=1))
    my_sorted_ind = np.argsort(my_dist)

    ant_per_viewing = {
        angle: my_ant[my_sorted_ind[(idx * 8):((idx + 1) * 8)]]
        for idx, angle in enumerate(VIEWING_ANGLES)
    }

    # vxvxB arm
    pos_vvB = angle_between(my_pos[my_sorted_ind, :2], np.array([0, 1])) < 1e-1
    neg_vvB = np.abs(np.pi - angle_between(my_pos[my_sorted_ind, :2], np.array([0, 1]))) < 1e-1
    tot_vvB = np.logical_or(pos_vvB, neg_vvB)

    ant_per_viewing_vvB = {
        angle: my_ant[my_sorted_ind][tot_vvB][(idx * 2):((idx + 1) * 2)]
        for idx, angle in enumerate(VIEWING_ANGLES)
    }

    for angle, ants in ant_per_viewing_vvB.items():
        print('=====')
        print(angle)
        print('----')
        for ant in ants:
            print(f'{calculate_viewing_angle(my_slice, my_shower, ant):.3f}')
        print('-----')
        for ant in ants:
            print(f'{calculate_viewing_angle_ant(my_slice, my_shower, ant):.3f}')


    cmap = mpl.colormaps.get_cmap('viridis')
    norm = mpl.colors.Normalize(vmin=min(VIEWING_ANGLES), vmax=max(VIEWING_ANGLES))

    # plt.scatter(my_pos[:, 0], my_pos[:, 1], c='k')
    # for angle, ants in ant_per_viewing_vvB.items():
    #     for ant in ants:
    #         plt.scatter(
    #             *my_shower.ant_position_showerplane(ant)[:2], color=cmap(norm(angle))
    #         )
    # plt.show()

    # Evaluate spectral functions
    delta_xmax_origin = GRAMMAGE - my_shower.xmax
    delta_xmax_target = GRAMMAGE - my_shower_2.xmax
    freq_range = np.logical_and(
        my_synthesis.frequency_range[0] <= my_synthesis.frequencies,
        my_synthesis.frequencies <= my_synthesis.frequency_range[1]
    )

    geo_origin, ce_origin, ce_lin_origin = my_slice.get_spectra(
        delta_xmax_origin, my_synthesis.frequencies[freq_range] - my_synthesis.frequency_range[2]
    )
    geo_target, ce_target, ce_lin_target = my_slice.get_spectra(
        delta_xmax_target, my_synthesis.frequencies[freq_range] - my_synthesis.frequency_range[2]
    )

    geo_factor, ce_factor = my_slice.get_normalisation_factors(my_shower)
    for angle in ant_per_viewing_vvB.keys():
        fig, ax = plt.subplots(2, 2, figsize=(14, 10))
        for ant in ant_per_viewing_vvB[angle]:
            trace_slice_geo, trace_slice_ce = my_shower.get_trace_slice(ant)
            trace_slice_geo_2, trace_slice_ce_2 = my_shower_2.get_trace_slice(ant)
            distance = calculate_distance_to_ant(my_slice, my_shower, ant)

            geo_corr = 2 * distance / geo_factor / (units.microvolt / units.m)
            ax[0, 0].plot(
                trace_slice_geo * geo_corr, label=ant
            )
            ax[0, 0].plot(
                trace_slice_geo_2 * geo_corr, '--', label=ant
            )
            ax[1, 0].plot(
                my_synthesis.frequencies / units.MHz, np.abs(np.fft.rfft(trace_slice_geo, norm='ortho')) * geo_corr,
                label=ant
            )
            ax[1, 0].plot(
                my_synthesis.frequencies / units.MHz, np.abs(np.fft.rfft(trace_slice_geo_2, norm='ortho')) * geo_corr,
                '--', label=ant
            )

            ce_corr = 2 * distance / ce_factor / (units.microvolt / units.m)
            ax[0, 1].plot(
                trace_slice_ce * ce_corr, label=ant
            )
            ax[0, 1].plot(
                trace_slice_ce_2 * ce_corr, '--', label=ant
            )
            ax[1, 1].plot(
                my_synthesis.frequencies / units.MHz, np.abs(np.fft.rfft(trace_slice_ce, norm='ortho')) * ce_corr,
                label=ant
            )
            ax[1, 1].plot(
                my_synthesis.frequencies / units.MHz, np.abs(np.fft.rfft(trace_slice_ce_2, norm='ortho')) * ce_corr,
                '--', label=ant
            )

        ax[1, 0].plot(
            my_synthesis.frequencies[freq_range] / units.MHz, geo_origin[matching_index_viewing_angles[angle]], 'k'
        )
        ax[1, 0].plot(
            my_synthesis.frequencies[freq_range] / units.MHz, geo_target[matching_index_viewing_angles[angle]], 'k--'
        )

        ax[1, 1].plot(
            my_synthesis.frequencies[freq_range] / units.MHz, ce_lin_origin[matching_index_viewing_angles[angle]], 'k'
        )
        ax[1, 1].plot(
            my_synthesis.frequencies[freq_range] / units.MHz, ce_lin_target[matching_index_viewing_angles[angle]], 'k--'
        )

        ax[0, 0].set_xlim([50, 1000])
        ax[0, 1].set_xlim([50, 1000])
        ax[1, 0].set_xlim([0, 500])
        ax[1, 1].set_xlim([0, 500])

        ax[1, 1].legend()

        ax[0, 0].set_title('Geomagnetic component')
        ax[0, 1].set_title('Charge-excess component')

        fig.suptitle(
            f'Angle = {angle} \n'
            f'Solid Delta = {delta_xmax_origin:.3f} \n'
            f'Dashed Delta = {delta_xmax_target:.3f}'
        )

        fig.savefig(f'output/Compare_spectra_{my_shower.name}_{my_shower_2.name}_{GRAMMAGE}_{int(angle * 1000)}.png',
                    bbox_inches='tight', dpi=200)

        plt.close(fig)



import jax
import jax.numpy as jnp


# source: https://stackoverflow.com/questions/2827393/angles-between-two-n-dimensional-vectors-in-python
def unit_vector(vector : jax.typing.ArrayLike) -> jax.Array:
    """ Returns the unit vector of the vector.  """
    return vector / jnp.linalg.norm(vector, axis=0)


def angle_between(v1 : jax.typing.ArrayLike, v2 : jax.typing.ArrayLike) -> jax.Array:
    """ Returns the angle in radians between vectors 'v1' and 'v2' """
    if v1.shape[-1] != 3 or v2.shape[-1] != 3:
        raise ValueError("Lengths dont match. Last axis should be position")
    v1_u = unit_vector(v1.T).T
    v2_u = unit_vector(v2.T).T
    return jnp.arccos(jnp.clip(jnp.einsum("...i,...i-> ...", v1_u, v2_u), -1.0, 1.0))



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'SMIET'
copyright = '2025, Mitja Desmet, Keito Watanabe'
author = 'Mitja Desmet, Keito Watanabe'
release = '0.8'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx_immaterial",
    "sphinx.ext.todo",
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
]

templates_path = ['_templates']
exclude_patterns = []

# Display todos by setting to True
# todo_include_todos = True

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_immaterial'
html_static_path = ['_static']
html_theme_options = {
    "icon": {
        "repo": "fontawesome/brands/github",
        "edit": "material/file-edit-outline",
    },
    "palette": [
        {
            "media": "(prefers-color-scheme: light)",
            "scheme": "default",
            "primary": "light-green",
            "accent": "light-blue",
            "toggle": {
                "icon": "material/lightbulb",
                "name": "Switch to dark mode",
            },
        },
        {
            "media": "(prefers-color-scheme: dark)",
            "scheme": "slate",
            "primary": "deep-orange",
            "accent": "lime",
            "toggle": {
                "icon": "material/lightbulb-outline",
                "name": "Switch to light mode",
            },
        },
    ],
    "site_url": "https://kwat0308.github.io/smiet-docs/"
}


from .synthesis import TemplateSynthesis
from .io import SlicedShower
from .io import BaseShower as Shower
from .io import CoreasShower

__all__ = [
    "Shower",
    "SlicedShower",
    "CoreasShower",
    "TemplateSynthesis",
]



import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors

from cr_pulse_interpolator import signal_interpolation_fourier as sigF
from cr_pulse_interpolator import interpolation_fourier as interpF

from smiet import units
from smiet.numpy import geo_ce_to_e, SlicedShower, TemplateSynthesis

# Selections (to be put in main)
F_MIN, F_MAX, F_0 = np.array([30, 80, 50]) * units.MHz
INTERPOLATION = 'trace'


# Convenience functions
def rot_matrix(angle):
    matrix = np.array([[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]])
    return matrix


# Variables
shower_path = '/home/mitjadesmet/Data/ShowersSKA/OriginShowers'
shower_origin = 130165
shower_target = 130188

origin = SlicedShower(f'{shower_path}/SIM{shower_origin}.hdf5')
target = SlicedShower(f'{shower_path}/SIM{shower_target}.hdf5')

synthesis = TemplateSynthesis()
synthesis.read_spectral_file(
    f'spectral_parameters_{int(F_MIN / units.MHz)}_{int(F_MAX / units.MHz)}_{int(F_0 / units.MHz)}.hdf5'
)
synthesis.make_template(origin)

total_geo_synth, total_ce_synth = synthesis.map_template(target)

transformer = origin.get_transformer()

x, y, z = transformer.transform_to_vxB_vxvxB(synthesis.antenna_information['position'], core=origin.core).T

# Check starshape
radius = np.linspace(20.0, 100.0, 5)
radius_ext = np.linspace(125.0, 225.0, 5)
radius = np.concatenate((radius, radius_ext))

fig, ax = plt.subplots(1, 1, figsize=(10, 10))
for r in radius[:-1]:
    # Starshape of SKA simulations is off, but still on concentric circles
    circle = plt.Circle((0, 0), r, color='r', fill=False)
    ax.add_patch(circle)
ax.scatter(x, y, s=5)
ax.set_aspect('equal')

plt.show()

# Convert GEO/CE back to on-sky
on_sky_star_shape = geo_ce_to_e(np.stack((total_geo_synth, total_ce_synth), axis=2), x, y)
footprint_star_shape = np.zeros((*total_geo_synth.shape, 3))
for ind in range(footprint_star_shape.shape[0]):
    footprint_star_shape[ind] = transformer.transform_from_ground_to_onsky(
        on_sky_star_shape[ind].T
    ).T

# Make signal interpolator
interpolator = sigF.interp2d_signal(
    x, y,
    footprint_star_shape,
    lowfreq=F_MIN / units.MHz,
    highfreq=F_MAX / units.MHz,
    verbose=True,
    phase_method='phasor',
    sampling_period=target.coreas_settings['time_resolution'] / units.s
)

# SKA shower set is rotated, so generate normal starshape from interpolator
one_arm_distances = np.concatenate((
    [12.5 / 8, 12.5 / 4, 12.5 / 2],
    np.arange(12.5, 200, 12.5),
    np.arange(200, 300, 25)
))

starshape_pos_x = []
starshape_pos_y = []
for angle in [0, 45, 90, 135, 180, 225, 270, 315]:
    x, y = np.matmul(rot_matrix(np.deg2rad(angle)), np.stack((one_arm_distances, np.zeros_like(one_arm_distances))))
    starshape_pos_x.extend(x)
    starshape_pos_y.extend(y)

starshape_pos_x = np.array(starshape_pos_x)
starshape_pos_y = np.array(starshape_pos_y)

starshape_values = np.zeros((len(starshape_pos_x), *footprint_star_shape.shape[1:]))
for ind in range(len(starshape_pos_x)):
    starshape_values[ind, :, :] = interpolator(starshape_pos_x[ind], starshape_pos_y[ind])

starshape_fluences = np.sum(starshape_values ** 2, axis=(1, 2))

# And now we make a fluence interpolator to check the footprint
fluence_interpolator = interpF.interp2d_fourier(starshape_pos_x, starshape_pos_y, starshape_fluences)

# Calculate footprint
dist_scale = 300.0
ti = np.linspace(-dist_scale, dist_scale, 1000)
XI, YI = np.meshgrid(ti, ti)

ZI = fluence_interpolator(XI, YI)
maxp = np.max(ZI)

# Plot footprint
norm = mcolors.Normalize(
    vmin=0,
    vmax=maxp
)
cmap = mpl.colormaps.get_cmap('jet')

fig, ax = plt.subplots(1, 1, figsize=(10, 10))

ax.pcolor(XI, YI, ZI, cmap=cmap, norm=norm)
ax.scatter(starshape_pos_x, starshape_pos_y, marker='+', s=3, color='w')

mm = mpl.cm.ScalarMappable(cmap=cmap)
mm.set_array([0.0, maxp])

cbar = fig.colorbar(mm, ax=ax)
cbar.set_label('Values of f(x, y)')

ax.set_xlabel('x [ m ]')
ax.set_ylabel('y [ m ]')
ax.set_xlim(-300, 300)
ax.set_ylim(-300, 300)
ax.set_aspect('equal')

plt.savefig(f'Footprint_synthesis_{shower_origin}_{shower_target}.png')

print("Done")



"""
This script was originally written by Tim Huege and Felix Schlüter, and is adapted
to store HDF5 files from origin showers, which additionally applies a bandpass filter
and resamples the traces to 2 GHz sampling.

Additionally, it is possible to only store antennas at the v x v x B axis, which is useful
if we only need to work on vxvxB arm instead.
"""

from __future__ import division

#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import numpy as np
import configparser
import sys
import io
import re
import argparse
import h5py
from scipy.signal import resample_poly


VERSION_MAJOR = 0
VERSION_MINOR = 5

def bandpass_filter_trace(trace, trace_sampling, f_min, f_max, sample_axis=0):
    """
    Bandpass filter a trace between `f_min` and `f_max`. The `trace` array can be multidimensional, in
    which case the `sample_axis` parameter indicates which dimension should be taken as the time
    samples (ie this parameter is passed on to the `np.fft.rfft` call).

    NB: taken from smiet.numpy.utilities

    Parameters
    ----------
    trace : np.ndarray
        The array containing the time traces to be filtered
    trace_sampling : float
        The sampling interval of the time traces, in internal units
    f_min : float
        The lower frequency to filter by, in internal units
    f_max : float
        The upper frequency to filter by, in internal units
    sample_axis : int, default=0
        The axis of `trace` which contains the time samples

    Returns
    -------
    filtered_trace : np.ndarray
        The filtered traces, in the same shape as `trace`

    Notes
    -----
    To avoid issues when the maximum of the trace is too close to the edge, all traces are first
    shifted to have their maxima more or less in the middle. After the filter has been applied,
    the traces are rolled back so that they are on same time axis as the input traces.
    """
    # Assuming `trace_sampling` has the correct internal unit, freq is already in the internal unit system
    freq = np.fft.rfftfreq(trace.shape[sample_axis], d=trace_sampling)
    freq_range = np.logical_and(freq > f_min, freq < f_max)

    # Find the median maximum sample number of the traces
    max_index = np.median(np.argmax(trace, axis=sample_axis))
    to_roll = int(trace.shape[sample_axis] / 2 - max_index)

    # Roll all traces such that max is in the middle
    roll_pulse = np.roll(trace, to_roll, axis=sample_axis)

    # FFT, filter, IFFT
    spectrum = np.fft.rfft(roll_pulse, axis=sample_axis)
    spectrum = np.apply_along_axis(
        lambda ax: ax * freq_range.astype("int"), sample_axis, spectrum
    )
    filtered = np.fft.irfft(spectrum, axis=sample_axis)

    return np.roll(filtered, -to_roll, axis=sample_axis)

def get_antenna_position_showerplane(cstrafo, antenna_position : np.ndarray, core : np.ndarray) -> np.ndarray:
    """Get the antenna position in the shower plane"""
    # modify the coordinate system of the antenna positions first
    ant_pos = np.copy(antenna_position)
    ant_pos[0] = -1 * ant_pos[1]  # north -> east
    ant_pos[1] = antenna_position[0]  # west -> north
    ant_pos[2] = antenna_position[2]  # height -> height
    return cstrafo.transform_to_vxB_vxvxB(
        ant_pos, core=core
    )

def get_transformer(f_h5_inputs):
    """
    Get the coordinate transformation object.

    The coordinate transformation object is obtained through radiotools,
    where the inputs from corsika are first converted into the Auger / NRR
    coordinate system.

    Parameters
    ----------
    f_h5_inputs : h5py.Group
        The hdf5 group containing the corsika input parameters.
    """
    Bx, Bz = f_h5_inputs.attrs["MAGNET"]
    zenith = np.deg2rad(f_h5_inputs.attrs["THETAP"][0])
    # shift the azimuth to get to Auger / NRR coordinate system
    azimuth = 3 * np.pi / 2. + np.deg2rad(f_h5_inputs.attrs["PHIP"][0])
    azimuth = rdhelp.get_normalized_angle(azimuth)
    B_inclination = np.arctan2(Bz, Bx)
    B_declination = 0
    # as in Auger / NRR coordinate system, we shift the declination by 90 deg
    magnetic_field_vector = rdhelp.spherical_to_cartesian(B_inclination + np.pi / 2, B_declination + np.pi * 0.5)

    ctrans = coordinatesystems.cstrafo(zenith, azimuth, magnetic_field_vector=magnetic_field_vector)

    return ctrans

def read_input_file(hdf5_file, inp_file):
    if not isinstance(inp_file, io.IOBase):
        inp_file = open(inp_file, "r")

    inp_dict = {}
    for i, line in enumerate(inp_file.readlines()):
        elements = line.strip().split()
        inp_dict[elements[0]] = elements[1:]
    inp_file.close()

    f_h5_inputs = hdf5_file.create_group("inputs")

    # fill general attributes from inp file
    f_h5_inputs.attrs["RUNNR"] = int(inp_dict["RUNNR"][0])
    f_h5_inputs.attrs["EVTNR"] = int(inp_dict["EVTNR"][0])
    f_h5_inputs.attrs["PRMPAR"] = int(inp_dict["PRMPAR"][0])
    f_h5_inputs.attrs["ERANGE"] = np.array(
        [float(inp_dict["ERANGE"][0]), float(inp_dict["ERANGE"][1])]
    )
    f_h5_inputs.attrs["THETAP"] = np.array(
        [float(inp_dict["THETAP"][0]), float(inp_dict["THETAP"][1])]
    )
    f_h5_inputs.attrs["PHIP"] = np.array(
        [float(inp_dict["PHIP"][0]), float(inp_dict["PHIP"][0])]
    )
    f_h5_inputs.attrs["ECUTS"] = np.array(
        [
            float(inp_dict["ECUTS"][0]),
            float(inp_dict["ECUTS"][1]),
            float(inp_dict["ECUTS"][2]),
            float(inp_dict["ECUTS"][3]),
        ]
    )
    try:
        f_h5_inputs.attrs["THIN"] = np.array(
            [
                float(inp_dict["THIN"][0]),
                float(inp_dict["THIN"][1]),
                float(inp_dict["THIN"][2]),
            ]
        )
        f_h5_inputs.attrs["THINH"] = np.array(
            [float(inp_dict["THINH"][0]), float(inp_dict["THINH"][1])]
        )
    except KeyError:
        pass
    f_h5_inputs.attrs["OBSLEV"] = float(inp_dict["OBSLEV"][0])
    f_h5_inputs.attrs["MAGNET"] = np.array(
        [float(inp_dict["MAGNET"][0]), float(inp_dict["MAGNET"][1])]
    )

    try:
        f_h5_inputs.attrs["ATMOD"] = int(inp_dict["ATMOD"][0])
    except KeyError:
        f_h5_inputs.attrs["ATMOD"] = 1  # CORSIKA default, U.S standard by Linsley

    if "ATMFILE" in inp_dict:
        atm_entry = str(inp_dict["ATMFILE"][0])
        print("ATMFILE was set in *inp, store ATMOD = -1")
        f_h5_inputs.attrs["ATMFILE"] = atm_entry
        f_h5_inputs.attrs["ATMOD"] = -1


def read_reas_file(hdf5_file, reas_file):
    if not isinstance(reas_file, io.IOBase):
        reas_file = open(reas_file, "r")

    lreas = reas_file.readlines()
    tmp = "[CoREAS]\n"
    for i, line in enumerate(lreas):
        lreas[i] = line.strip().split(";")[0].strip()
        tmp += lreas[i] + "\n"
    reas_file.close()

    configParser = configparser.RawConfigParser()
    configParser.optionxform = str
    tmp2 = io.StringIO(tmp)
    configParser.read_file(tmp2)
    items = configParser.items("CoREAS")

    f_h5_reas = hdf5_file.create_group("CoREAS")

    # store content of reas file as attributes
    for key, value in items:
        if key in [
            "CoreCoordinateNorth",
            "CoreCoordinateWest",
            "CoreCoordinateVertical",
            "TimeResolution",
            "AutomaticTimeBoundaries",
            "TimeLowerBoundary",
            "TimeUpperBoundary",
            "ResolutionReductionScale",
            "GroundLevelRefractiveIndex",
            "EventNumber",
            "RunNumber",
            "GPSSecs",
            "GPSNanoSecs",
            "CoreEastingOffline",
            "CoreNorthingOffline",
            "CoreVerticalOffline",
            "RotationAngleForMagfieldDeclination",
            "ShowerZenithAngle",
            "ShowerAzimuthAngle",
            "PrimaryParticleEnergy",
            "PrimaryParticleType",
            "DepthOfShowerMaximum",
            "DistanceOfShowerMaximum",
            "MagneticFieldStrength",
            "MagneticFieldInclinationAngle",
        ]:
            if key not in [
                "EventNumber",
                "RunNumber",
                "GPSSecs",
                "GPSNanoSecs",
                "PrimaryParticleType",
            ]:
                f_h5_reas.attrs[key] = float(value)
            else:
                f_h5_reas.attrs[key] = int(value)
        else:
            f_h5_reas.attrs[key] = value


def read_longitudinal_profile(hdf5_file, long_file):
    if not isinstance(long_file, io.IOBase):
        long_file = io.open(long_file, "r", encoding="UTF-8")

    f_h5_long = hdf5_file.create_group("atmosphere")

    lines = long_file.readlines()
    n_steps = int(lines[0].rstrip().split()[3])

    def my_replace(m):
        return str(m.group(0)[0]) + " -"

    n_data_str = io.StringIO()
    n_data_str.writelines(lines[2 : (n_steps + 2)])
    n_data_str.seek(0)

    n_data = np.genfromtxt(n_data_str)
    for i, line in enumerate(lines):
        lines[i] = re.sub("[0-9]-", my_replace, line)

    dE_data_str = io.StringIO()
    dE_data_str.writelines(lines[(n_steps + 4) : (2 * n_steps + 4)])
    dE_data_str.seek(0)
    dE_data = np.genfromtxt(dE_data_str)
    data_set = f_h5_long.create_dataset("NumberOfParticles", n_data.shape, dtype="f")
    data_set[...] = n_data
    data_set.attrs["comment"] = (
        "The collumns of the data set are: DEPTH, GAMMAS, POSITRONS, ELECTRONS, MU+, MU-, HADRONS, CHARGED, NUCLEI, CHERENKOV"
    )

    data_set = f_h5_long.create_dataset("EnergyDeposit", dE_data.shape, dtype="f")
    data_set[...] = dE_data
    data_set.attrs["comment"] = (
        "The collumns of the data set are: DEPTH, GAMMA, EM IONIZ, EM CUT, MU IONIZ, MU CUT, HADR IONIZ, HADR CUT, NEUTRINO, SUM"
    )

    # read out hillas fit
    hillas_parameter = []
    for line in lines:
        if bool(re.search("PARAMETERS", line)):
            hillas_parameter = [
                float(x) for x in line.split()[2:]
            ]  # strip aways 'PARAMETER', '='
    f_h5_long.attrs["Gaisser-Hillas-Fit"] = hillas_parameter

    long_file.close()


def read_atm_file(hdf5_file, atm_file_path):
    heights, refractive_index = np.genfromtxt(atm_file_path, unpack=True, skip_header=6)
    refractive_index_profile = np.array([heights, refractive_index]).T

    f_h5_atm = hdf5_file.create_group("atmosphere_model")

    data_set = f_h5_atm.create_dataset(
        "RefractiveIndexProfile", refractive_index_profile.shape, dtype="f"
    )
    data_set[...] = refractive_index_profile

    atm_file = open(atm_file_path, "rb")
    lines = atm_file.readlines()

    # skip first entry (0), conversion cm -> m
    h = np.array(lines[1].strip(b"\n").split()[1:], dtype=float) / 100
    a = np.array(lines[2].strip(b"\n").split(), dtype=float) * 1e4
    b = np.array(lines[3].strip(b"\n").split(), dtype=float) * 1e4
    c = np.array(lines[4].strip(b"\n").split(), dtype=float) * 1e-2
    atm_file.close()

    f_h5_atm.attrs["h"] = h
    f_h5_atm.attrs["a"] = a
    f_h5_atm.attrs["b"] = b
    f_h5_atm.attrs["c"] = c


def read_antenna_data(
    hdf5_file, list_file, antenna_folder, cstrafo, core_pos, store_only_vxvxB: bool = False, trace_params : dict = {
        "f_min": 30e6,
        "f_max": 500e6,
        "dt_resample" : 1e-9,
        "upsampling_factor": 5
    },
):
    if not isinstance(list_file, io.IOBase):
        list_file = open(list_file, "r")

    lines = [item for item in list_file.readlines() if item != "\n"]  # skip empty lines
    lines = np.array(
        [item for item in lines if not item.startswith("#")]
    )  # skip comments
    list_file.close()

    dt_original = hdf5_file.attrs["TimeResolution"]
    dt_resample = trace_params["dt_resample"]  
    upsampling_factor = trace_params["upsampling_factor"] 
    downsampling_factor = int(upsampling_factor / (dt_original / dt_resample))
    print("Resampling traces from %.1f to dt = %.1f ns (up: %d, down: %d)" % (dt_original * 1e9, dt_resample * 1e9, upsampling_factor, downsampling_factor))

    observers = hdf5_file.create_group("observers")

    for line in lines:
        ll = line.strip().split()
        antenna_position = np.array(ll[2:5], dtype=float)
        antenna_label = ll[5]

        # get position in shower plane
        ant_pos_showerplane = get_antenna_position_showerplane(cstrafo, antenna_position / 100, core_pos)  # convert to m

        if store_only_vxvxB:
            # find the vxvxB coordinate
            # it should be ~ 0 in vxB axis and > 0 in vxvxB axis
            ants_at_vxvxB = (np.abs(ant_pos_showerplane[0]) < 1) & (np.sign(ant_pos_showerplane[1]) > 0)
            if not ants_at_vxvxB:
                # print("Skipping antenna %s, not at vxvxB axis." % antenna_label)
                continue

            # print("Reading antenna %s at position %s m (shower plane: %s m)" % (antenna_label, antenna_position / 100, ant_pos_showerplane))

        antenna_file = os.path.join(antenna_folder, "raw_%s.dat" % antenna_label)

        data = np.genfromtxt(antenna_file)

        # at this point, already bandpass and downsample
        trace_time, trace_val = data[:, 0], data[:, 1:4]

        # NB: everything is in seconds, so we convert all relevant
        # quantities to seconds
        filtered_trace = bandpass_filter_trace(
            trace_val,
            trace_sampling=dt_original,
            f_min=trace_params["f_min"],  # 30 MHz
            f_max=trace_params["f_max"],  # 500 MHz
            sample_axis=0
        )
        
        downsampled_trace = resample_poly(filtered_trace, up=upsampling_factor, down=downsampling_factor, axis=0)

        # also convert the timing
        n_resampled = downsampled_trace.shape[0]
        trace_time_resampled = np.arange(n_resampled) * dt_resample + trace_time[0]
        # combine the data back
        filtered_resampled_data = np.concatenate(
            (trace_time_resampled[:, np.newaxis], downsampled_trace), axis=-1
        )
        data_set = observers.create_dataset(antenna_label, filtered_resampled_data.shape, dtype=float, compression='gzip')
        data_set[...] = filtered_resampled_data

        data_set.attrs["position"] = np.array(antenna_position, dtype=float)
        data_set.attrs["name"] = antenna_label

        # there seems to be a problem when storing a list of unicodes in an attribute (is the case for python3). followed the fix from:
        # https://github.com/h5py/h5py/issues/289
        if len(ll) > 6:
            data_set.attrs["additional_arguments"] = [a.encode("utf8") for a in ll[6:]]

    # modify the timing resolution
    hdf5_file.attrs["TimeResolution"] = dt_resample


def write_coreas_hdf5_file(
    reas_filename, output_filename, f_h5=None, atm_file=None, ignore_atm_file=False, only_vxvxB=False
):
    # create hdf5 file
    if f_h5 is None:
        f_h5 = h5py.File(
            output_filename, "w", driver="core"
        )  # "core": will only write file to disk ones it is closed
    f_h5.attrs["version_major"] = VERSION_MAJOR
    f_h5.attrs["version_minor"] = VERSION_MINOR

    # read all parameter from the SIM??????.reas file
    read_reas_file(f_h5, reas_filename)

    # print error message in case CorsikaParameterFile was not specified in the .reas file
    inp_filename = f_h5["CoREAS"].attrs["CorsikaParameterFile"]
    if len(inp_filename) == 0:
        sys.exit("CorsikaParameterFile was not specified in the .reas file, aborting!")
    finp = open(os.path.join(os.path.dirname(reas_filename), inp_filename), "r")

    # read cards from .inp file and stores them into "inputs" group in f_h5 file
    read_input_file(f_h5, finp)
    print("Read input parameters from: %s" % inp_filename)

    if "ATMFILE" in f_h5["inputs"].attrs:
        print("Found following path for ATMFILE: %s" % f_h5["inputs"].attrs["ATMFILE"])
        atm_file_full_path = f_h5["inputs"].attrs["ATMFILE"]
        atm_file_name = os.path.basename(atm_file_full_path)

        if atm_file is not None:
            print("Use user specified ATMFILE: %s" % atm_file)
            read_atm_file(f_h5, atm_file)
        elif os.path.exists(atm_file_full_path):
            print("Read ATMFILE from: %s" % atm_file_full_path)
            read_atm_file(f_h5, atm_file_full_path)
        elif os.path.exists(atm_file_name):
            print("Read ATMFILE from: %s" % atm_file_name)
            read_atm_file(f_h5, atm_file_name)
        elif ignore_atm_file:
            print("Do not read any ATMFILE")
            pass
        else:
            sys.exit(
                "Could not find the atm file in %s or %s. Full stop!"
                % (atm_file_full_path, atm_file_name)
            )

    # read in long file
    long_filename = (
        "DAT" + os.path.splitext(os.path.basename(reas_filename))[0][-6:] + ".long"
    )
    longinp = io.open(
        os.path.join(os.path.dirname(reas_filename), long_filename),
        "r",
        encoding="UTF-8",
    )
    read_longitudinal_profile(f_h5, longinp)
    print("Read longitudinal profile from: %s" % long_filename)

    # read in antenna data
    listfile = open(os.path.splitext(reas_filename)[0] + ".list", "r")
    number = os.path.splitext(os.path.basename(reas_filename))[0][3:]
    antenna_folder = os.path.join(
        os.path.dirname(reas_filename), "SIM%s_coreas" % number
    )
    cstrafo = get_transformer(f_h5["inputs"])
    core_pos = np.array(
        [
            -1 * f_h5["CoREAS"].attrs["CoreCoordinateWest"],
            f_h5["CoREAS"].attrs["CoreCoordinateNorth"],
            f_h5["inputs"].attrs["OBSLEV"],
        ]
    ) / 100  # convert to m
    read_antenna_data(f_h5["CoREAS"], listfile, antenna_folder, cstrafo, core_pos, only_vxvxB)
    print("Read antenna data from folder: %s" % antenna_folder)

    return f_h5
    

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="coreas hdf5 converter for origin showers")
    parser.add_argument("input_file", type=str, help="path to input reas file")

    parser.add_argument(
        "-o",
        "--outputDirectory",
        dest="output_directory",
        type=str,
        default=None,
        help="path to output hdf5 directory",
    )
    parser.add_argument(
        "-of",
        "--outputFileName",
        dest="output_filename",
        type=str,
        default=None,
        help="output filename",
    )

    parser.add_argument(
        "--only_vxvxB",
        action="store_true",
        help="Store only the v x v x B polarization component of the electric field (Default: false).",
    )

    parser.add_argument(
        "--ignoreATMfile",
        action="store_true",
        help="If a ATMFILE was used for the simulation but can"
        " not be found choose to ignore it (Default: false).",
    )

    parser.add_argument(
        "--atmfile",
        type=str,
        default=None,
        help='Path to a GDAS atm file. Only required if "ATMFILE" specified in *inp file.'
        " If not provided check for file and location specified in *inp file.",
    )

    args = parser.parse_args()

    reas_filename = args.input_file
    output_basename = os.path.splitext(os.path.basename(reas_filename))[0]
    if args.only_vxvxB:
        output_basename += "_vxvxB"
    output_filename = output_basename + ".hdf5"

    if args.output_filename is not None:
        output_filename = args.output_filename
    if args.output_directory is not None:
        output_filename = os.path.join(args.output_directory, output_filename)

    try:
        from radiotools import helper as rdhelp
        from radiotools import coordinatesystems
    except ModuleNotFoundError as e:
        sys.exit("Could not find the radiotools module: '{}'\n"
                    "You can get this module from https://github.com/nu-radio/radiotools.\n"
                    "Make sure to add it to your enviourment, e.g., PYTHONPATH too. Stopping ...".format(e))

    f_h5 = write_coreas_hdf5_file(
        reas_filename,
        output_filename,
        atm_file=args.atmfile,
        ignore_atm_file=args.ignoreATMfile,
        only_vxvxB=args.only_vxvxB,
    )
    f_h5.close()



import os
import logging
import argparse
import subprocess


logger = logging.getLogger("smiet.make_docs")


if __name__ == '__main__':
    argparser = argparse.ArgumentParser(
        prog="smiet.docs.make_docs",
        description=(
            "A script to automatically generate the API documentation for template synthesis,"
            " and build the html documentation for online publishing."
        )
    )
    argparser.add_argument(
        '--no-clean', default=False, const=True, action='store_const',
        help=(
            "Do not delete existing html files and build only pages which have changed."
            " Useful if you are only modifying or adding (not moving/removing) pages.")
    )
    argparser.add_argument(
        '--skip-apidoc', default=False, const=True, action='store_const', help=(
            "Skip the compilation of the automatic API documentation. "
            "Speeds up the compilation of the documentation if no changes were made to the code.")
    )
    parsed_args = argparser.parse_args()

    # create the automatic code documentation with apidoc
    output_folder = 'source/apidoc'
    if os.path.exists(output_folder):
        if not parsed_args.no_clean: # remove old apidoc folder
            logger.info('Removing old apidoc folder: {}'.format(output_folder))
            subprocess.check_output(['rm', '-rf', output_folder])


    module_path = '../smiet/'
    subprocess.run(
        [
            'sphinx-apidoc', '-efMT', '-d', '1', '--ext-autodoc', '--ext-intersphinx',
            '--ext-coverage', '--ext-githubpages', '-o', output_folder,
            module_path
        ]
    )

    # We don't use the top level APIDOC toctrees, so we remove them to eliminate a sphinx warning
    subprocess.check_output([
        'rm',
        os.path.join(output_folder, 'smiet.numpy.rst'),
        os.path.join(output_folder, 'smiet.jax.rst'),
        os.path.join(output_folder, 'smiet.rst')
    ])

    if os.path.exists('build'):
        subprocess.check_output(['rm', '-rf', 'build'])
    subprocess.run(
            ['sphinx-build', '-b', 'html', 'source', 'build/html'],
        )



import logging

import h5py
import numpy as np

from typing_extensions import Self, Tuple, Union

from ..utilities import e_to_geo_ce, bandpass_filter_trace
from .coreas_shower import CoreasShower
from smiet import units


class SlicedShower(CoreasShower):
    """
    This class can be used to read in an HDF5 file of sliced CoREAS simulation.

    It can read in the traces of all slices for a given antenna and return them in a Numpy array.
    Note that the implementation assumes that the antenna's in each slice have the same physical position
    and are named identically apart from a suffix 'xN', where N is the grammage at the **bottom** of the slice.
    The antenna names are inferred by splitting the antenna names on the 'x' character, taking the first part
    and adding them all to a set to removes doubles.
    The number of slices is then calculated as the number of observers in the file, divided by the number of
    unique antenna names (i.e. the length of the set).
    Finally, the antenna positions are added to a structured array by looping over all unique antenna names,
    adding `x{slice_grammage}` (of which the value is inferred from the longitudinal profiel) at the end and
    looking at the position attribute. This structured array is then sorted by distance to the core, to ensure
    consistency across read-ins.

    Parameters
    ----------
    file_path : str
        The path to the HDF5 file
    gdas_file : str, optional
        GDAS file to be used for the Atmosphere object. See :obj:`smiet.numpy.coreas_shower.CoreasShower`
        for more info.

    Attributes
    ----------
    antenna_names : set
        The set of all antenna names present in this shower
    antenna_array : np.ndarray
        A structured array containing for each antenna the name and position on ground.
        It contains the fields 'name' and 'position'.
    nr_slices : int
        The number of slices in the shower
    """

    def __init__(self, file_path, gdas_file=None):
        # Set self.antenna_names to something other than None, to avoid reading in the antennas in the parents init
        self.antenna_names = set()
        super().__init__(file_path, gdas_file=gdas_file)

        self.logger = logging.getLogger("smiet.numpy.io.SlicedShower")

        with h5py.File(self._file) as file:
            self._read_antennas(file)

            # Do not use the GH parameter from Shower fit, but rather those from the file
            # in order: [N, X0, Xmax, p0, p1, p2] (p0,p1,p2 are the polynomial coefficients in denominator for lambda)
            GH_param = file["atmosphere"].attrs["Gaisser-Hillas-Fit"]
            lmbda = np.mean(
                    GH_param[3]
                    + GH_param[4] * self.grammages
                    + GH_param[5] * self.grammages**2)
            
            self.GH_parameters = {
                "nmax": GH_param[0],
                "X0": GH_param[1],
                "xmax": GH_param[2],
                "lambda": lmbda,
                "L" : np.sqrt(np.abs((GH_param[2] - GH_param[1]) * lmbda)),
                "R" : np.sqrt(np.abs(lmbda / (GH_param[2] - GH_param[1])))
            }

            # Count number of slices
            self.nr_slices = len(file["CoREAS"]["observers"].keys()) // len(
                self.antenna_names
            )

    def _read_antennas(self, file):
        self.logger.info(f"Reading antennas from {file.name}...")

        # Extract antenna information
        self.antenna_names = set(
            [key.split("x")[0] for key in file["CoREAS"]["observers"].keys()]
        )

        # Trace length can be longer than coreas settings
        self._trace_length = len(
            file["CoREAS"]["observers"][
                f"{next(iter(self.antenna_names))}x{int(self.slice_grammage)}"
            ]
        )

        # Create the antenna array, which is a structured array containing the name and position of each antenna
        self._make_antenna_array(
            {
                ant_name: file["CoREAS/observers"][
                    f"{ant_name}x{int(self.slice_grammage)}"
                ].attrs["position"]
                * units.cm
                for ant_name in self.antenna_names
            }
        )

        # And sort the antennas, such that the order is always the same
        self._sort_antenna_array()

    def get_trace_geoce(
        self: Self,
        ant_name: str,
        bandpass: list = [30 * units.MHz, 500 * units.MHz],
        return_start_time: bool = False,
    ) -> Union[
        Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray, np.ndarray]
    ]:
        """
        Get the traces from all slices for a given antenna, in GEO/CE components.

        Notes
        -----
        The traces are converted to GEO/CE components.

        Parameters
        ----------
        ant_name : str
            The name of the antenna. Must be the same as the key in the HDF5!
        bandpass : list, default=[30, 500]
            If not None, the traces are bandpass filtered between the given frequencies (in MHz)
        return_start_time : bool, default=False
            If True, an array containing the time of the first sample of each slice is returned

        Returns
        -------
        traces_geo : np.ndarray
            The geomagnetic traces, shaped as (slices, samples)
        traces_ce : np.ndarray
            The charge-excess traces, shaped as (slices, samples)
        traces_start_times : np.ndarray (returned only if return_start_times is True)
            The time of the first sample of each trace
        """
        if ant_name not in self.antenna_names:
            raise ValueError(f"Antenna name {ant_name} is not present in shower")

        traces_geo = np.zeros((self.nr_slices, self._trace_length))
        traces_ce = np.zeros((self.nr_slices, self._trace_length))
        traces_start_times = np.zeros((self.nr_slices,))
        for i_slice in range(self.nr_slices):
            g_slice = (i_slice + 1) * int(self.slice_grammage)
            traces_geo[i_slice], traces_ce[i_slice], traces_start_times[i_slice] = (
                self.get_trace_slice_geoce(
                    ant_name, g_slice, bandpass, return_start_time=True
                )
            )

        if return_start_time:
            return traces_geo, traces_ce, traces_start_times

        return traces_geo, traces_ce

    def get_trace_slice_geoce(self, ant_name, slice_value, bandpass, return_start_time=False):
        """
        Similar to :func:`get_trace`, but only gets the traces for a single slice in GEO/CE components.

        Parameters
        ----------
        slice_value : int
            The grammage of the slice. Must exist in the HDF5.
        ant_name : str
            The name of the antenna. Must be the same as the key in the HDF5!
        return_start_time : bool, default=False
            If True, an array containing the time of the first sample of each slice is returned

        Returns
        -------
        trace_geo : np.ndarray
            The geomagnetic trace
        trace_ce : np.ndarray
            The charge-excess trace
        trace_start_time : np.ndarray (returned only if return_start_times is True)
            The time of the first sample of the trace
        """
        if ant_name not in self.antenna_names:
            raise ValueError(f"Antenna name {ant_name} is not present in shower")

        transformer = self.get_transformer()
        trace_slice_ground, trace_time_axis = self.get_trace_raw(
            f"{ant_name}x{slice_value}"
        )
        trace_slice_filtered = bandpass_filter_trace(
            trace_slice_ground,
            self.coreas_settings["time_resolution"],
            bandpass[0],
            bandpass[1],
            sample_axis=-1,
        )
        trace_slice_vvB = transformer.transform_to_vxB_vxvxB(trace_slice_filtered).T

        # unit of pos does not matter, this is divided away
        antenna_vvB = self.get_antenna_position_showerplane(ant_name)
        trace_geo, trace_ce = e_to_geo_ce(trace_slice_vvB, *antenna_vvB[:2])

        if return_start_time:
            return trace_geo, trace_ce, trace_time_axis[0]

        return trace_geo, trace_ce

    def get_trace_slice_on_sky(self, ant_name, slice_value, return_start_time=False):
        """
        Similar to :func:`get_trace_slice`, but the traces are converted to on-sky components
        instead of GEO/CE components.

        Parameters
        ----------
        slice_value : int
            The grammage of the slice. Must exist in the HDF5.
        ant_name : str
            The name of the antenna. Must be the same as the key in the HDF5!
        return_start_time : bool, default=False
            If True, an array containing the time of the first sample of each slice is returned

        Returns
        -------
        trace_geo : np.ndarray
            The geomagnetic trace
        trace_ce : np.ndarray
            The charge-excess trace
        trace_start_time : np.ndarray (returned only if return_start_times is True)
            The time of the first sample of the trace
        """
        if ant_name not in self.antenna_names:
            raise ValueError(f"Antenna name {ant_name} is not present in shower")

        transformer = self.get_transformer()
        trace_slice_ground, trace_time_axis = self.get_trace_raw(
            f"{ant_name}x{slice_value}"
        )
        trace_slice_on_sky = transformer.transform_from_ground_to_onsky(
            trace_slice_ground
        ).T

        if return_start_time:
            return trace_slice_on_sky, trace_time_axis[0]

        return trace_slice_on_sky


class SlicedShowerCherenkov(CoreasShower):
    """
    This class can be used to read in an HDF5 file of sliced CoREAS simulation, where every slice has
    the same number of antennas.

    It differs from :obj:`SlicedShower` in that it does not assume that all observers have the same
    position. Rather, it only assumes that each slice has the same number of observers.
    The antenna names are taken to be all observer names in HDF5 (they are still stored in a set,
    but this should not change anything because the observer names in CoREAS should be unique).
    The ``antenna_array`` is created by looping over all antenna names and extracting the position attribute.
    It is also sorted by distance to the core, to ensure it is always the same when reading in a shower.
    To calculate the number of slices, it divides the number of observers by the number of observers
    whose name ends with the slice grammage extracted from the longitudinal profile.

    Its primary use case is for sliced simulation where the antennas where all place under the same
    Cherenkov angle in each slice.

    Parameters
    ----------
    file_path : str
        The path to the HDF5 file
    gdas_file : str, optional
        GDAS file to be used for the Atmosphere object. See :obj:`smiet.numpy.coreas_shower.CoreasShower`
        for more info.

    Attributes
    ----------
    antenna_names : set
        The set of all antenna names present in this shower
    antenna_array : np.ndarray
        A structured array containing for each antenna the name and position on ground.
        It contains the fields 'name' and 'position'.
    nr_slices : int
        The number of slices in the shower
    """

    def __init__(self, file_path, gdas_file=None):
        super().__init__(file_path, gdas_file=gdas_file)

        self.logger = logging.getLogger("smiet.numpy.io.SlicedShowerCherenkov")

        with h5py.File(self._file) as file:
            # Do not use the GH parameter from Shower fit, but rather those from the file
            self._GH_parameters = file["atmosphere"].attrs["Gaisser-Hillas-Fit"]

            # Count number of slices
            self.nr_slices = len(file["CoREAS"]["observers"].keys()) // len(
                [
                    el
                    for el in file["CoREAS"]["observers"].keys()
                    if el.split("x")[1] == str(int(self.slice_grammage))
                ]
            )

    def get_trace_slice(self, ant_name, slice_gram=None, return_start_time=False):
        """
        Retrieves the traces for an observer and converts them to GEO/CE components.

        Parameters
        ----------
        ant_name : str
            The name of the observer as used in the HDF5 file
        slice_gram : None
            This variable  is not used, but is kept to mirror the method of SlicedShower,
            which allows this class to be used in TemplateSynthesis
        return_start_time : bool, default=False
            If True, an array containing the time of the first sample of each slice is returned

        Returns
        -------
        trace_geo : np.ndarray
            The geomagnetic trace
        trace_ce : np.ndarray
            The charge-excess trace
        trace_start_time : np.ndarray (returned only if return_start_times is True)
            The time of the first sample of the trace
        """
        if ant_name not in self.antenna_names:
            raise ValueError(f"Antenna name {ant_name} is not present in shower")

        transformer = self.get_transformer()
        trace_slice_ground, trace_time_axis = self.get_trace_raw(f"{ant_name}")
        trace_slice_showerplane = transformer.transform_to_vxB_vxvxB(
            trace_slice_ground
        ).T

        # unit of pos does not matter, this is divided away
        antenna_showerplane = self.get_antenna_position_showerplane(ant_name)
        trace_geo, trace_ce = e_to_geo_ce(
            trace_slice_showerplane, *antenna_showerplane[:2]
        )

        if return_start_time:
            return trace_geo, trace_ce, trace_time_axis[0]

        return trace_geo, trace_ce



import numpy as np


def bandpass_filter_trace(trace, trace_sampling, f_min, f_max, sample_axis=0):
    """
    Bandpass filter a trace between `f_min` and `f_max`. Both should be provided in the internal unit
    system, just like the `trace_sampling` parameter. The `trace` array can be multidimensional, in
    which case the `sample_axis` parameter indicates which dimension should be taken as the time
    samples (ie this parameter is passed on to the `np.fft.rfft` call).

    Parameters
    ----------
    trace : np.ndarray
        The array containing the time traces to be filtered
    trace_sampling : float
        The sampling interval of the time traces, in internal units
    f_min : float
        The lower frequency to filter by, in internal units
    f_max : float
        The upper frequency to filter by, in internal units
    sample_axis : int, default=0
        The axis of `trace` which contains the time samples

    Returns
    -------
    filtered_trace : np.ndarray
        The filtered traces, in the same shape as `trace`

    Notes
    -----
    To avoid issues when the maximum of the trace is too close to the edge, all traces are first
    shifted to have their maxima more or less in the middle. After the filter has been applied,
    the traces are rolled back so that they are on same time axis as the input traces.
    """
    # Assuming `trace_sampling` has the correct internal unit, freq is already in the internal unit system
    trace_length = trace.shape[sample_axis]
    freq = np.fft.rfftfreq(trace_length, d=trace_sampling)
    freq_range = np.logical_and(freq > f_min, freq < f_max)

    # Find the median maximum sample number of the traces
    max_index = np.median(np.argmax(trace, axis=sample_axis))
    to_roll = int(trace_length / 2 - max_index)

    # Roll all traces such that max is in the middle
    roll_pulse = np.roll(trace, to_roll, axis=sample_axis)

    # FFT, filter, IFFT
    spectrum = np.fft.rfft(roll_pulse, axis=sample_axis, n=trace_length)
    spectrum = np.apply_along_axis(
        lambda ax: ax * freq_range.astype("int"), sample_axis, spectrum
    )
    filtered = np.fft.irfft(spectrum, axis=sample_axis, n=trace_length)

    return np.roll(filtered, -to_roll, axis=sample_axis)


def transform_traces_on_vxB(my_geo, my_ce, x, y, vxB_axis_mode="replace"):
    """
    Adjust the (synthesised) traces on the vxB axis.

    This function is meant to alter the traces on the vxB axis, as in the current
    version of the SMIET software the vxB axis cannot be synthesised. There are several
    strategies to handle this, and applying them should be facilitated by this function.

    Parameters
    ----------
    my_geo: np.ndarray
        The electric field traces of the geomagnetic component, shaped as (NANTS, NSAMPLES)
    my_ce: np.ndarray
        The electric field traces of the charge-excess component, shaped as (NANTS, NSAMPLES)
    x : np.ndarray
        the x-coordinates of the antennas in the vxB axis, shaped as (NANTS,).
    y : np.ndarray
        the y-coordinates of the antennas in the vxvxB axis, shaped as (NANTS,).
    vxB_axis_mode : {"replace", "zero"}
        Operation to perform on the vxB axis. If "replace" (the default), the traces from the
        vxvxB component will be copied over to the vxB component. If "zero", the vxB component
        will be set to zero.

    Returns
    -------
    my_geo_adjusted : np.ndarray
        The adjusted geomagnetic component traces, shaped as (NANTS, NSAMPLES)
    my_ce_adjusted : np.ndarray
        The adjusted charge-excess component traces, shaped as (NANTS, NSAMPLES)
    """
    ant_at_vB_axis = np.abs(y) < 1e-3
    ant_at_vvB_axis = np.abs(x) < 1e-3

    if vxB_axis_mode == "replace":
        # for the vxB component, replace with the vxvxB component
        my_geo[ant_at_vB_axis, :] = my_geo[ant_at_vvB_axis, :]
        my_ce[ant_at_vB_axis, :] = my_ce[ant_at_vvB_axis, :]
    elif vxB_axis_mode == "zero":
        # for the vxB component, set to zero
        my_geo[ant_at_vB_axis, :] = 0.0
        my_ce[ant_at_vB_axis, :] = 0.0
    else:
        raise ValueError(
            f"Unknown vxB_axis_mode: {vxB_axis_mode}. Must be one of 'replace' or 'zero'."
        )

    return my_geo, my_ce


def reconstruct_starshape_from_vxvxB(
    geo_traces, ce_traces, antenna_positions_vvB, number_of_arms=8
):
    """
    Reconstruct traces on all arms from vxvxB-arm simulation (where geo and ce are orthogonal).

    This is used for generating star-shapes from simulations that only have antennas on the
    vxvxB arm. The traces on the other arms are simply copied from the vxvxB arm, and the
    antenna positions are rotated accordingly.

    This can also be passed traces synthesised for all slices. In that case, the output traces will
    have the same number of slices. If only at the ground, the n_slices = 1.

    Parameters
    ----------
    geo_traces : np.ndarray
        Geomagnetic traces on the vxvxB arm, shape (n_antennas, n_samples, n_slices)
    ce_traces : np.ndarray
        Charge-excess traces on the vxvxB arm, shape (n_antennas, n_samples, n_slices)
    antenna_positions_vvB : np.ndarray
        Positions of antennas on vxvxB arm (distance in shower plane), shape (n_antennas,)
    number_of_arms : int, default=8
        Total number of star-shape arms to reconstruct

    Returns
    -------
    geo_traces_starshape : np.ndarray
        Geomagnetic traces on all arms, shape (n_antennas*number_of_arms, n_samples, n_slices)
    ce_traces_starshape : np.ndarray
        Charge-excess traces on all arms, shape (n_antennas*number_of_arms, n_samples, n_slices)
    antenna_positions_starshape : np.ndarray
        Antenna positions on all arms, shape (n_antennas*number_of_arms, 3)
    """
    n_ant_per_arm = geo_traces.shape[0]

    # get the total number of antennas
    n_ant = n_ant_per_arm * number_of_arms

    # initialise arrays
    ant_labels = []
    antenna_positions_starshape = np.zeros((n_ant, 3))
    geo_traces_starshape = np.zeros((n_ant, *geo_traces.shape[1:]))
    ce_traces_starshape = np.zeros((n_ant, *geo_traces.shape[1:]))

    phi_angles = np.linspace(0, 360, number_of_arms, endpoint=False)

    for ip, phi in enumerate(phi_angles):
        phi_rad = np.deg2rad(phi)
        geo_traces_starshape[ip * n_ant_per_arm : (ip + 1) * n_ant_per_arm, :] = (
            geo_traces
        )
        ce_traces_starshape[ip * n_ant_per_arm : (ip + 1) * n_ant_per_arm, :] = (
            ce_traces
        )
        antenna_positions_starshape[
            ip * n_ant_per_arm : (ip + 1) * n_ant_per_arm, 0
        ] = antenna_positions_vvB * np.cos(phi_rad)
        antenna_positions_starshape[
            ip * n_ant_per_arm : (ip + 1) * n_ant_per_arm, 1
        ] = antenna_positions_vvB * np.sin(phi_rad)

        ant_labels += [f"ant_r{antenna_positions_vvB[i]:.0f}_phi{phi:.0f}" for i in range(n_ant_per_arm)]

    return geo_traces_starshape, ce_traces_starshape, antenna_positions_starshape, ant_labels



import logging
from typing_extensions import Self, Tuple, Union

import numpy as np
import h5py
import jax

jax.config.update("jax_enable_x64", True)

import jax.numpy as jnp
from scipy.constants import c as c_vacuum

from jax_radio_tools import units, trace_utils, shower_utils
from jax_radio_tools.atmosphere import Atmosphere
from jax_radio_tools.cstrafo_jax import (
    e_to_geo_ce,
    conversion_fieldstrength_cgs_to_SI,
    hp_get_normalized_angle,
)
from .base_shower import BaseShower

logger = logging.getLogger("smiet.jax.io")


class CoreasShower(BaseShower):
    """
    Class to read in showers from CoREAS simulations stored in HDF5 files.

    Parameters
    ----------
    file_path : str
        Path to the HDF5 file containing the CoREAS simulation data.
    bandpass : list, default = [30 * units.MHz, 500 * units.MHz]
        The frequency bandpass in MHz to apply to the traces. Defaults to 30-500 MHz,
        which is the bandwidth in which SMIET is applicable in.
    gdas_file : str, optional
        If provided, the atmosphere will be created using this GDAS file. This will take
        precedence over the atmosphere model specified in the simulation settings.
    """

    def __init__(
        self: Self,
        file_path: str,
        bandpass: list = [30 * units.MHz, 500 * units.MHz],
        gdas_file: str = "",
    ) -> None:
        super().__init__()
        self.__file = file_path  # input file path

        self.__trace_length = None
        self.__delta_t = None
        self.__frequency_range = bandpass
        self.name = file_path.split("/")[-1].split(".")[0]

        self.antenna_array = {"name": None, "position": None, "dis_to_core" : None}

        self.traces = None
        self.trace_times = None

        self.__parse_hdf5(gdas_file=gdas_file)

        # filter traces
        self.traces = trace_utils.filter_trace(
            self.traces,
            f_min=self.__frequency_range[0],
            f_max=self.__frequency_range[1],
            trace_sampling=self.delta_t,
            sample_axis=2,
        )

    def __parse_hdf5(
        self: Self,
        gdas_file: str = "",
    ) -> None:
        file = h5py.File(self.__file)

        # set up longitudinal profile parameters
        self._setup_long_parameters(file)

        # now also setup the corsika parameters
        self._setup_corsika_parameters(file)

        # after this, setup the atmosphere
        # imperative that this is done after setting up the core & zenith angle
        self._setup_atmosphere(file, gdas_file=gdas_file)

        # load the antenna and trace information here
        # this should be called after setting the grammage and long profile
        self._load_antenna_and_traces(file)

        # sortthe antenna array to increasing distance from core & azimuth
        self._sort_antenna_array()

        file.close()

    def _setup_long_parameters(self: Self, file: h5py.File) -> None:
        """
        Setup parameters from the longitudinal profile.
        """
        # longitudinal profile data
        long_data = file["atmosphere"]["NumberOfParticles"][:]

        self.grammages = long_data[:, 0]
        self.long_profile = np.sum(long_data[:, 2:4], axis=1)

        # Gaisser-Hillas parameters
        # in particular used to set nmax and xmax
        # in order: [N, X0, Xmax, p0, p1, p2] (p0,p1,p2 are the polynomial coefficients in denominator for lambda)
        GH_parameters = np.array(
            file["atmosphere"].attrs["Gaisser-Hillas-Fit"]
        )  # GH parameters from fit
        self.nmax = GH_parameters[0]
        self.xmax = GH_parameters[2]

        # convert GH parameter to LR
        lmbda = np.mean(
            GH_parameters[3]
            + GH_parameters[4] * self.grammages
            + GH_parameters[5] * self.grammages**2
        )
        L, R = shower_utils.convert_GH_to_LR(self.xmax, GH_parameters[1], lmbda)

        self.GH_parameters = {
            "x0": GH_parameters[1],
            "lambda": lmbda,
            "L": L,
            "R": R,
            "xmax": self.xmax,
            "nmax": self.nmax,
        }

    def _setup_corsika_parameters(self: Self, file: h5py.File) -> None:
        """
        Setup parameters specific to CORSIKA simulations.
        """
        # get (EM) energy
        self.energy = file["inputs"].attrs["ERANGE"][0] * units.GeV
        self.primary = file["inputs"].attrs["PRMPAR"]  # integer stored for convenience

        self.zenith = np.deg2rad(file["inputs"].attrs["THETAP"][0]) * units.rad
        # Shift the azimuthal angle by 3 pi/2 (180 flip from West -> East plus N-> E due to convention of second coordinate)
        self.azimuth = (
            hp_get_normalized_angle(
                3 * np.pi / 2 + np.deg2rad(file["inputs"].attrs["PHIP"][0])
            )
            * units.rad
        )

        # # 3. define magnetic field vector
        self.magnet = np.array([
            0, file["inputs"].attrs["MAGNET"][0], -1 * file["inputs"].attrs["MAGNET"][1]
        ]) * units.microtesla
        # 4. get simulation core
        self.core = (
            np.array(
                [
                    -1 * file["CoREAS"].attrs["CoreCoordinateWest"],
                    file["CoREAS"].attrs["CoreCoordinateNorth"],
                    file["CoREAS"].attrs["CoreCoordinateVertical"],
                ]
            )
            * units.cm
        )
        self.core[2] = file["inputs"].attrs["OBSLEV"] * units.m

        self.delta_t = float(file["CoREAS"].attrs["TimeResolution"]) * units.s

    def _setup_atmosphere(self: Self, file: h5py.File, gdas_file: str = "") -> None:
        """Setup the atmosphere object, and truncate grammages and longitudinal profile if needed."""
        # if the ATMOD value is -1, then use gdas file as atmosphere
        if file["inputs"].attrs["ATMOD"] == -1:
            self.atm = Atmosphere(
                observation_level=self.core[2] / units.m, gdas_file=gdas_file
            )
        else:
            self.atm = Atmosphere(
                model=file["inputs"].attrs["ATMOD"],
                observation_level=self.core[2] / units.m,
            )

        # filter out grammages that will reach beyond earth
        grammage_lim_mask = self.grammages <= self.atm.get_xmax_from_distance(
            0.0,
            self.zenith / units.rad,
        )

        self.long_profile = self.long_profile[grammage_lim_mask]
        self.grammages = self.grammages[grammage_lim_mask]

    def _load_antenna_and_traces(self: Self, file: h5py.File) -> None:
        """Read in the antenna & trace information from the HDF5 file."""
        # antenna positions at the ground
        self.antenna_array["name"] = list(file["CoREAS"]["observers"].keys())
        self.antenna_array["position"] = np.array(
            [
                file["CoREAS"]["observers"][f"{ant_name}"].attrs["position"] * units.cm
                for ant_name in self.antenna_array["name"]
            ]
        )  # antennas x 3
        self.antenna_array["position"] = self.antenna_array["position"][:, [1, 0, 2]]
        self.antenna_array["position"][:,0] *= -1

        # distance from the shower core
        self.antenna_array["dis_to_core"] = np.linalg.norm(
            (self.get_antenna_position_showerplane()[:, :2]), axis=-1
        )

        # the observer data containing both timing and trace information
        observers = np.array(
            [
                file["CoREAS"]["observers"][f"{ant_name}"][:]
                for ant_name in self.antenna_array["name"]
            ]
        )  # antennas x samples x 4

        # map it to the correct shape of 4 x ANT x SAMPLES
        observers = np.moveaxis(observers, (0, 1, 2), (1, 2, 0))

        # divide into trace times & actual E-field slices
        self.trace_times = observers[0, ...] * units.s  # (Nant, Nsamples)
        self.traces = observers[1:, ...]  # (Npol, Nant, Nsamples)
        self.trace_length = self.traces.shape[2]

        self.traces = self.traces[[1, 0, 2], ...] * conversion_fieldstrength_cgs_to_SI  # convert traces from cgs to SI manually (due to electric charge shenanigans)
        self.traces[0,...] *= -1     

    def _sort_antenna_array(self: Self) -> None:
        """Sort the trace and antenna positions based on increasing distance from the core and azimuth."""
        # 7. finally sort based on increasing distance to core & increasing azimuth.
        # first sort based on increasing distance
        dis_core_mask = np.argsort(self.antenna_array["dis_to_core"])
        self.antenna_array["position"] = self.antenna_array["position"][
            dis_core_mask, :
        ]
        self.antenna_array["name"] = [
            self.antenna_array["name"][i] for i in dis_core_mask
        ]
        self.traces = self.traces[:, dis_core_mask, ...]
        self.trace_times = self.trace_times[dis_core_mask, :]
        self.antenna_array["dis_to_core"] = np.sort(self.antenna_array["dis_to_core"])
        # now sort it based on azimuth for each distance
        # NOTE: key assumption applied where it is sorted based on the fact that there are 8 arms
        # when we introduce interpolator we would not need this anymore
        n_ant_per_arm = len(self.antenna_array["dis_to_core"]) // 8

        # calculate the angles
        angles = np.arctan2(
            self.get_antenna_position_showerplane()[:, 1],
            self.get_antenna_position_showerplane()[:, 0],
        )
        angles = np.around(angles, 15)
        angles[angles < 0] += 2 * np.pi
        # get the indices in increasing order and shift by n_arms
        # then concatenate to get the indices
        ordered_idces = [
            np.argsort(angles[int(8 * iant) : int(8 * (iant + 1))]) + 8 * iant
            for iant in range(n_ant_per_arm)
        ]
        ordered_idces = np.array(np.concatenate(ordered_idces), dtype=int)

        # finally apply to the arrays
        self.antenna_array["position"] = self.antenna_array["position"][
            ordered_idces, :
        ]
        self.antenna_array["name"] = [
            self.antenna_array["name"][i] for i in ordered_idces
        ]
        self.traces = self.traces[:, ordered_idces, ...]
        self.trace_times = self.trace_times[ordered_idces, :]
        self.antenna_array["dis_to_core"] = self.antenna_array["dis_to_core"][ordered_idces]

    @property
    def trace_length(self: Self) -> int:
        """
        Length of the trace.

        Returns
        -------
        trace_length : int
            The number of samples in the trace.
        """
        return self.__trace_length

    @property
    def delta_t(self: Self) -> float:
        """
        Time resolution of the trace.

        Returns
        -------
        delta_t : float
            The time resolution in ns.
        """
        return self.__delta_t

    @trace_length.setter
    def trace_length(self: Self, trace_length: int) -> None:
        self.__trace_length = trace_length

    @delta_t.setter
    def delta_t(self: Self, delta_t: float) -> None:
        self.__delta_t = delta_t

    def get_antenna_position(self: Self) -> np.ndarray:
        """
        Get the antenna positions on the ground.

        Returns
        -------
        ant_position_ground : np.ndarray
            The antenna positions on the ground, transformed to the NRR system.
        """
        return self.antenna_array["position"]

    def get_antenna_position_showerplane(self: Self) -> np.ndarray:
        """
        Get the antenna positions in shower plane.

        Returns
        -------
        ant_position_vB_vvB : jax.typing.ArrayLike
            The antenna positions in the shower plane, transformed to the NRR system.
        """
        return self.get_transformer().transform_to_vxB_vxvxB(
            self.antenna_array["position"], core=self.core
        )

    def get_traces_raw(self: Self) -> np.ndarray:
        """
        Get all traces for all antennas on the ground plane.

        Returns
        -------
        traces_ground : jax.typing.ArrayLike
            the traces at ground in xyz coordinates.
            Shape is (3, Nant, Nsamples), where Nant is the number of antennas and Nsamples is the number of samples.

            When called with SlicedShower, the shape is (3, Nant, Nsamples, Nslices).
        """
        return self.traces

    def get_traces_onsky(self: Self) -> np.ndarray:
        """
        Get the traces from traces at the ground to on-sky components.

        NOTE: the radial component (er) is set to zero manually here.

        Returns
        -------
        traces_onsky : np.ndarray
            The traces in the shower plane, transformed to on-sky components (er, etheta, ephi).
            Shape is (3, Nant, Nsamples), where Nant is the number of antennas and Nsamples is the number of samples.

            When called with SlicedShower, the shape is (3, Nant, Nsamples, Nslices).
        """
        trace_onsky = (
            self.get_transformer()
            .transform_from_ground_to_onsky(
                self.traces.T,
            )
            .T
        )
        trace_onsky[0, ...] = 0.0  # set er component to zero
        return trace_onsky

    def get_traces_vB_vvB(self: Self) -> np.ndarray:
        """
        Get the traces from traces at the ground to vB/vvB components.

        Returns
        -------
        traces_vB_vvB : np.ndarray
            The traces in the shower plane, transformed to vxB and vxvxB components.
            Shape is (3, Nant, Nsamples), where Nant is the number of antennas and Nsamples is the number of samples.

            When called with SlicedShower, the shape is (3, Nant, Nsamples, Nslices).
        """
        trace_vvB = (
            self.get_transformer()
            .transform_to_vxB_vxvxB(self.traces.T, core=self.core)
            .T
        )
        return trace_vvB

    def get_traces_geoce(self: Self) -> np.ndarray:
        """
        Get the traces from traces at the ground to GEO/CE components.

        Returns
        -------
        traces_geo_ce : np.ndarray
            The traces in the shower plane, transformed to GEO and CE components.
            Shape is (2, Nant, Nsamples), where Nant is the number of antennas and Nsamples is the number of samples.

            When called with SlicedShower, the shape is (2, Nant, Nsamples, Nslices).
        """
        traces_vvB = self.get_traces_vB_vvB()
        ant_position_vvB = self.get_antenna_position_showerplane()
        # now convert trace to respective components
        traces_geo, traces_ce = e_to_geo_ce(
            traces_vvB.T, ant_position_vvB[None, :, 0], ant_position_vvB[None, :, 1]
        )
        return np.array([traces_geo.T, traces_ce.T])

    def remove_antennas(self, ant_idx_min: int = 0, ant_idx_max: int = 90) -> None:
        """
        Remove antennas from the data traces.

        Parameter:
        ----------
        ant_idx_min : float
            remove all antennas below this antenna index
        ant_idx_max : float
            remove all antennas above this antenna index
        """
        # d_core_mask = (self.dis_to_core > d_core_min) & (self.dis_to_core < d_core_max)
        d_core_mask = (np.indices(self.dis_to_core.shape)[0] >= ant_idx_min) & (
            np.indices(self.dis_to_core.shape)[0] < ant_idx_max
        )
        self.traces = self.traces[:, d_core_mask, ...]
        self.trace_times = self.trace_times[d_core_mask, :]
        self.antenna_array["position"] = self.antenna_array["position"][d_core_mask, :]
        self.antenna_array["name"] = [
            self.antenna_array["name"][i]
            for i in range(len(self.antenna_array["name"]))
            if d_core_mask[i]
        ]
        self.dis_to_core = self.dis_to_core[d_core_mask]



from . import write_simulation
from .write_simulation import *

__all__ = []
__all__ += write_simulation.__all__

from .generate_origin_library import OriginLibraryGenerator

__all__.append("OriginLibraryGenerator")
